import uuid

import atexit
import os
import sys
from unittest import mock
import tornado.testing
import subprocess
from unittest.mock import patch

from tinybird.kafka_utils import KafkaUtils
from .views.base_test import BaseTest
import shutil

from tinybird import data_connector
from tinybird.user import UserAccount, User

KAFKA_PATH = os.environ.get('KAFKA_PATH', os.path.join(os.path.dirname(__file__), '../..', 'kafka_2.13-2.8.0'))


class TestKafka(BaseTest):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        shutil.rmtree('/tmp/kafka-logs-test', ignore_errors=True)
        shutil.rmtree('/tmp/zookeeper-kafka-test', ignore_errors=True)

        args = (
            f'{KAFKA_PATH}/bin/zookeeper-server-start.sh',
            os.path.join(os.path.dirname(__file__), '..', './tests/kafka-zookeeper-test.properties')
        )
        cls.zookeeper_process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=sys.stdout, universal_newlines=True)

        @atexit.register
        def kill_zk():
            cls.zookeeper_process.kill()

        found = False
        while not found:
            if cls.zookeeper_process.poll() is not None:
                break
            for line in iter(cls.zookeeper_process.stdout.readline, ''):
                if line.find('Using checkInterval') != -1:
                    print("ZK started!")
                    found = True
                    break
                if line.find('Unexpected exception') != -1:
                    print("Could not start ZK process: " + line.rstrip())
                    cls.assertTrue(False, None, msg="ZK process not started")
                    return

        args = (
            f'{KAFKA_PATH}/bin/kafka-server-start.sh',
            os.path.join(os.path.dirname(__file__), '..', './tests/kafka-test-server.properties')
        )
        cls.kafka_process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=sys.stdout, universal_newlines=True)

        @atexit.register
        def kill_kafka():
            cls.kafka_process.kill()

        found = False
        while not found:
            if cls.kafka_process.poll() is not None:
                break

            for line in iter(cls.kafka_process.stdout.readline, ''):
                if line.find('Recorded new controller') != -1:
                    print("KAFKA server started : " + line.rstrip())
                    found = True
                    break
                if line.find('Unexpected exception') != -1:
                    print("Could not start Kafka process: " + line.rstrip())
                    cls.assertTrue(False, None, msg="Kafka process not started")
                    return

        cls.zookeeper_process.stdout.close()
        cls.kafka_process.stdout.close()

    def setUp(self):
        super().setUp()
        self.assertEqual(self.zookeeper_process.poll(), None, msg="ZK process could not start")
        self.assertEqual(self.kafka_process.poll(), None, msg="Kafka process could not start")

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.kafka_process.kill()
        cls.zookeeper_process.kill()

    def produce_message(self, message, topic):
        producer_cmd = f"""{KAFKA_PATH}/bin/kafka-console-producer.sh --bootstrap-server localhost:9093 \
                            --topic {topic} --timeout 10 --producer-property enable.idempotence=true \
                            --request-required-acks -1 2>/dev/null"""

        c = os.system(f'''echo {message} | {producer_cmd}''')
        self.assertEqual(c, 0)

    def create_kafka_connector(self):
        user_account = UserAccount.register(f'test_kafka_preview{uuid.uuid4().hex}@mail', 'pass')
        workspace = self.register_workspace(f'test_kafka_preview_ws_{uuid.uuid4().hex}', user_account.id)

        conn = data_connector.DataConnector.add_connector(user=workspace, name="name", service='kafka', settings={
            'kafka_bootstrap_servers': 'localhost:9093',
            'kafka_security_protocol': 'plaintext',
            'kafka_sasl_plain_username': '',
            'kafka_sasl_plain_password': '',
        })

        return conn

    @tornado.testing.gen_test
    @patch('tinybird.data_connector.KafkaSettings.PREVIEW_POLL_TIMEOUT_MS', 2000)
    async def test_kafka_preview_all_params_happy(self):
        self.produce_message("""'{"test": 3}'""", "kafkafirst")
        conn = self.create_kafka_connector()
        r = await KafkaUtils.get_kafka_preview(conn,
                                               'test_groupid',
                                               kafka_topics=['kafkafirst'])
        assert r.get('error', None) is None
        output = r['preview']
        expected_deserialization = {
            'data': [{'test': 3}],
            'meta': [{'name': 'test', 'type': 'Int16'}],
            'rows': 1,
            'statistics': mock.ANY
        }
        self.assertEqual(len(output), 1)
        t = output[0]
        assert t['deserialized'] == expected_deserialization
        self.assertEqual(t['topic'], 'kafkafirst')
        self.assertEqual(t['messages_in_last_hour'], 1)
        last = t['last_messages']
        self.assertEqual(len(last), 1)
        self.assertEqual(last[0]['__partition'], 0)
        self.assertEqual(last[0]['__offset'], 0)
        self.assertEqual(last[0]['__key'], None)
        self.assertEqual(last[0]['__value'], '''{"test": 3}''')

    @tornado.testing.gen_test
    async def test_kafka_preview_topic_notexists(self):
        conn = self.create_kafka_connector()
        r = await KafkaUtils.get_kafka_preview(conn,
                                               'test_groupid',
                                               kafka_topics=['kafka_not_exists123'],
                                               preview_activity=True)
        self.assertEqual(r['error'], 'not_exists')

    @tornado.testing.gen_test
    @patch('tinybird.data_connector.KafkaSettings.PREVIEW_POLL_TIMEOUT_MS', 2000)
    async def test_kafka_preview_connector_id(self):
        self.produce_message("msg2", "kafkasecond")
        conn = self.create_kafka_connector()

        r = await KafkaUtils.get_kafka_preview(data_connector=conn, kafka_group_id='test_groupid', kafka_topics=['kafkasecond'])
        output = r['preview']
        self.assertEqual(len(output), 1)
        t = output[0]
        self.assertEqual(t['topic'], 'kafkasecond')
        self.assertEqual(t['messages_in_last_hour'], 1)
        self.assertEqual(t['committed'], False)
        last = t['last_messages']
        self.assertEqual(len(last), 1)
        self.assertEqual(last[0]['__partition'], 0)
        self.assertEqual(last[0]['__offset'], 0)
        self.assertEqual(last[0]['__key'], None)
        self.assertEqual(last[0]['__value'], "msg2")


class TestKafkaInvalidParams(BaseTest):
    @tornado.testing.gen_test
    @patch('tinybird.data_connector.KafkaSettings.PREVIEW_POLL_TIMEOUT_MS', 10)
    async def test_kafka_preview_all_params_invalid(self):
        user_account = UserAccount.register(f'test_kafka_preview{uuid.uuid4().hex}@mail', 'pass')
        workspace = self.register_workspace(f'test_kafka_preview_ws_{uuid.uuid4().hex}', user_account.id)

        conn = data_connector.DataConnector.add_connector(user=workspace, name="name", service='kafka', settings={
            'kafka_bootstrap_servers': 'localhost:49999',
            'kafka_security_protocol': 'plaintext',
            'kafka_sasl_plain_username': '',
            'kafka_sasl_plain_password': '',
        })

        output = await KafkaUtils.get_kafka_preview(conn,
                                                    'test_groupid')
        self.assertTrue('error' in output)


class TestGCloudScheduler(BaseTest):
    def test_create_data_connector_for_gcscheduler(self):
        connector_name = 'test_gcloud_connector'
        pipe_name = 'test_glcoud_connector_pipe'
        self.workspace = User.get_by_id(self.WORKSPACE_ID)
        copy_pipe = self.workspace.add_pipe(pipe_name, 'SELECT * FROM test_table')

        connector = data_connector.DataConnector.add_connector(
            user=self.workspace,
            name=connector_name,
            service=data_connector.DataConnectors.GCLOUD_SCHEDULER,
            settings={}
        )

        sink = data_connector.DataSink.add_sink(
            data_connector=connector,
            resource=copy_pipe,
            settings={}
        )

        self.assertEqual(sink.data_connector_id, connector.id)

        sink_by_pipe = data_connector.DataSink.get_by_pipe_id(copy_pipe.id)
        self.assertEqual(sink_by_pipe.id, sink.id)


class TestSnowflakeConnector(BaseTest):
    def test_create_snowflake_data_connector_fails_with_no_settings(self):
        connector_name = 'test_snowflake_connector'
        self.workspace = User.get_by_id(self.WORKSPACE_ID)
        with self.assertRaises(data_connector.InvalidSettingsException):
            _ = data_connector.DataConnector.add_connector(
                user=self.workspace,
                name=connector_name,
                service=data_connector.DataConnectors.SNOWFLAKE,
                settings={}
            )

    def test_create_snowflake_data_connector_succeeds_with_all_settings(self):
        connector_name = 'test_snowflake_connector'
        self.workspace = User.get_by_id(self.WORKSPACE_ID)
        settings = {
            'account': 'sf_account',
            'username': 'sf_username',
            'password': 'sf_password',
            'warehouse': 'sf_warehouse',
            'warehouse_size': 'X-Small',
            'role': 'sf_role'
        }
        connector = data_connector.DataConnector.add_connector(
            user=self.workspace,
            name=connector_name,
            service=data_connector.DataConnectors.SNOWFLAKE,
            settings=settings
        )

        self.assertEqual(connector.settings, settings)

    def test_create_snowflake_data_connector_succeeds_with_mandatory_settings(self):
        connector_name = 'test_snowflake_connector'
        self.workspace = User.get_by_id(self.WORKSPACE_ID)
        settings = {
            'account': 'sf_account',
            'username': 'sf_username',
            'password': 'sf_password'
        }
        connector = data_connector.DataConnector.add_connector(
            user=self.workspace,
            name=connector_name,
            service=data_connector.DataConnectors.SNOWFLAKE,
            settings=settings
        )

        self.assertEqual(connector.settings, settings)

    def test_create_snowflake_data_connector_fails_without_all_mandatory_settings(self):
        connector_name = 'test_snowflake_connector'
        self.workspace = User.get_by_id(self.WORKSPACE_ID)

        with self.assertRaises(data_connector.InvalidSettingsException):
            _ = data_connector.DataConnector.add_connector(
                user=self.workspace,
                name=connector_name,
                service=data_connector.DataConnectors.SNOWFLAKE,
                settings={'username': 'sf_username', 'password': 'sf_password'}
            )

        with self.assertRaises(data_connector.InvalidSettingsException):
            _ = data_connector.DataConnector.add_connector(
                user=self.workspace,
                name=connector_name,
                service=data_connector.DataConnectors.SNOWFLAKE,
                settings={'account': 'sf_account', 'password': 'sf_password'}
            )

        with self.assertRaises(data_connector.InvalidSettingsException):
            _ = data_connector.DataConnector.add_connector(
                user=self.workspace,
                name=connector_name,
                service=data_connector.DataConnectors.SNOWFLAKE,
                settings={'account': 'sf_account', 'username': 'sf_username'}
            )




from io import StringIO
import uuid
import tornado
import pytest

from tests.views.base_test import BaseTest, TBApiProxyAsync
from tests_e2e.aux import poll_async
from tinybird.constants import CHCluster
from tinybird.datasource_metrics import DataSourceMetrics

from tinybird.job import UsageMetricsTracker
from tinybird.token_scope import scopes
from tinybird.user import User, Users, public

from tests.conftest import is_main_process


@pytest.mark.serial  # Direct access to the public user
@pytest.mark.skipif(not is_main_process(), reason="Serial test")
class TestDataSourceMetrics(BaseTest):
    def setUp(self):
        super(TestDataSourceMetrics, self).setUp()
        self.original_user_get_all = User.get_all
        self.tb_api_proxy_async = TBApiProxyAsync(self)
        self.usage_metrics_tracker = UsageMetricsTracker()
        self.public_user = public.get_public_user()
        self.usage_metrics_storage = self.public_user.get_datasource('usage_metrics_storage__v2')
        self.interval = 60 * 60 * 24 * 7

    @tornado.testing.gen_test
    async def test_get_storage_metric(self):
        cluster = CHCluster('tinybird', User.default_database_server)
        workspace_name = f"test_get_total_rows_metric_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name, cluster)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(workspace)

        datasource_name = 'ds_test_get_storage_metric'
        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name=datasource_name,
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )
        datasource = Users.get_datasource(workspace, datasource_name)

        # wait for replication before tracking storage
        self.wait_for_datasource_replication(workspace, datasource)

        self.usage_metrics_tracker.track_usage_metrics_storage(
            self.public_user,
            [User.get_by_id(workspace.id)]
        )
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        metric_type = 'storage'
        datasource_metrics = DataSourceMetrics(workspace, datasource, metric_type, self.interval)
        metric = await datasource_metrics.get_metric()
        assert metric['data'][-1][metric_type] > 0
        assert metric['ticks'] == (self.interval / (60 * 60 * 24)) + 1

    @tornado.testing.gen_test
    async def test_get_total_rows_metric(self):
        cluster = CHCluster('tinybird', User.default_database_server)
        workspace_name = f"test_get_total_rows_metric_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name, cluster)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(workspace)
        datasource_name = 'ds_test_get_total_rows_metric'
        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name=datasource_name,
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )
        datasource = Users.get_datasource(workspace, datasource_name)

        # wait for replication before tracking storage
        self.wait_for_datasource_replication(workspace, datasource)

        self.usage_metrics_tracker.track_usage_metrics_storage(
            self.public_user,
            [User.get_by_id(workspace.id)]
        )
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        metric_type = 'total_rows'
        datasource_metrics = DataSourceMetrics(workspace, datasource, metric_type, self.interval)
        metric = await datasource_metrics.get_metric()
        assert metric['data'][-1][metric_type] == 2
        assert metric['ticks'] == (self.interval / (60 * 60 * 24)) + 1

    @tornado.testing.gen_test
    async def test_get_new_rows_metric(self):
        cluster = CHCluster('tinybird', User.default_database_server)
        workspace_name = f"test_get_new_rows_metric_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name, cluster)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(workspace)
        datasource_name = 'ds_test_get_new_rows_metric'
        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name=datasource_name,
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )
        datasource = Users.get_datasource(workspace, datasource_name)

        # wait for replication before tracking storage
        self.wait_for_datasource_replication(workspace, datasource)

        self.wait_for_public_table_replication('datasources_ops_log')
        metric_type = 'new_rows'
        datasource_metrics = DataSourceMetrics(workspace, datasource, metric_type, self.interval)

        async def f():
            metric = await datasource_metrics.get_metric()
            assert metric['data'][-1][metric_type] == 2
            assert metric['ticks'] == (self.interval / (60 * 60 * 24)) + 1

        await poll_async(f)

    @tornado.testing.gen_test
    async def test_get_new_rows_quarantine_metric(self):
        cluster = CHCluster('tinybird', User.default_database_server)
        workspace_name = f"test_get_new_rows_quarantine_metric_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name, cluster)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(workspace)
        datasource_name = 'ds_test_get_new_rows_metric'
        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name=datasource_name,
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )
        await self._insert_data_in_datasource(
            token=token,
            ds_name=datasource_name,
            data="1,2,3\n4,5,6\n7,8,9",
            assert_response=False
        )
        datasource = Users.get_datasource(workspace, datasource_name)

        # wait for replication before tracking storage
        self.wait_for_datasource_replication(workspace, datasource)

        self.wait_for_public_table_replication('datasources_ops_log')
        metric_type = 'new_rows_quarantine'
        datasource_metrics = DataSourceMetrics(workspace, datasource, metric_type, self.interval)

        async def f():
            metric = await datasource_metrics.get_metric()
            assert metric['data'][-1][metric_type] == 3
            assert metric['ticks'] == (self.interval / (60 * 60 * 24)) + 1

        await poll_async(f)



import math
import random
import pytest
import time
import tornado.testing

from statistics import mean, median, stdev
from tornado.ioloop import IOLoop

from tinybird.sql_template import get_template_and_variables
from tinybird.sql_toolset import replace_tables_chquery_cached, sql_get_used_tables_cached
from tinybird.token_scope import scopes
from tinybird.user import Users, User

from .conftest import is_main_process
from .views.base_test import BaseTest


def get_cache_info():
    hits, misses, _, _ = get_template_and_variables.cache_info()
    return hits, misses


def flush_all_caches():
    get_template_and_variables.cache_clear()
    sql_get_used_tables_cached.cache_clear()
    replace_tables_chquery_cached.cache_clear()


@pytest.mark.serial  # Better to test this serially (better to move it elsewhere too). It changes the cache status too
@pytest.mark.skipif(not is_main_process(), reason="Serial test")
class TestPerformance(BaseTest):
    ITERATIONS = 80

    def check_timings(self, timings, max_mean: float, max_p90: float, max_min: float):
        min_timings = min(timings)
        mean_timings = mean(timings)
        stdev_timings = stdev(timings)
        median_timings = median(timings)
        p90_timings = sorted(timings)[math.ceil(self.ITERATIONS * .9) - 1]
        print(f"""Timing metrics:
            min    {min_timings:.5f}
            mean   {mean_timings:.5f}
            stdev  {stdev_timings:.5f}
            median {median_timings:.5f}
            p90    {p90_timings:.5f}""")

        self.assertTrue(min_timings < max_min, f'Invalid min time performance: {min_timings} < {max_min}')
        self.assertTrue(mean_timings < max_mean, f'Invalid mean time performance: {mean_timings} < {max_mean}')
        self.assertTrue(p90_timings < max_p90, f'Invalid P90 time performance: {p90_timings} < {max_p90}')

    async def replace_tables_with_options(self, flush=True, randomize=None, sync=False):
        if flush:
            # Note that this only works correctly because we have only one runner
            await IOLoop.current().run_in_executor(User.replace_executor, flush_all_caches)

        if randomize:
            self.variables[randomize] = f"{random.randint(0, 1000000000)}"

        start = time.time()
        if sync:
            _ = self.u.replace_tables(f'SELECT * FROM {self.pipe.id}',
                                      readable_resources=self.readable_resources,
                                      pipe=self.pipe,
                                      use_pipe_nodes=False,
                                      variables=self.variables,
                                      template_execution_results=self.template_execution_results)
        else:
            _ = await self.u.replace_tables_async(f'SELECT * FROM {self.pipe.id}',
                                                  readable_resources=self.readable_resources,
                                                  pipe=self.pipe,
                                                  use_pipe_nodes=False,
                                                  variables=self.variables,
                                                  template_execution_results=self.template_execution_results)
        return time.time() - start

    async def get_replace_timings(self, flush=False, randomize=None):
        times = []
        for _i in range(self.ITERATIONS):
            t = await self.replace_tables_with_options(flush=flush, sync=False, randomize=randomize)
            times.append(t)
        return times

    @tornado.testing.gen_test(timeout=360)
    async def test_template_cache_random_parameter(self):

        datasources = [
            'article_commercial_tags_join_global_parent_partnumber',
            'articles_commercial_tags',
            'articles_filters',
            'articles_join',
            'articles_join_global_partnumber',
            'datetime_lut',
            'product_rank_month_mt',
            'product_rank_mt',
            'product_rank_rt_mt',
            'sales_historic_landing',
            'sales_historic_last_date',
            'sales_rt',
            'stores',
            'stores_join',
            'tagging_array',
            'worldwide_stores',
            'worldwide_stores_join_by_country',
        ]

        nodes = [
            {
                'name': 'all_stores',
                'sql': '''SELECT *
                FROM stores
                UNION ALL
                SELECT *
                FROM worldwide_stores '''
            },
            {
                'name': 'by_timestamp',
                'sql': '''%
                with
                    {% comment "prepare a map purchase_location -> offset to be used in query" %}
                    (
                        select length(location) > 0 ? location: [0] from (
                            select groupArray(location) location from (select * from stores order by cod_store)
                        )
                    ) as countries,
                    (
                        select length(offsets) > 0 ? offsets: [0] from (
                            select groupArray(-offset) offsets from
                            (select timezone from stores order by cod_store)
                            any left join
                            (select timezone, offset from datetime_lut where date = toDate(now()))
                            using timezone
                        )
                    ) as timezones
                SELECT
                cast(sku_rank_lc as String) sku_rank_lc,
                groupUniqArrayState(parent_catentry) as parent_catentry_array,
                {% for last, x in enumerate_with_last(split_to_array(attr, 'amount_return')) %}
                {% if x == 'order_count' %}
                        toInt64(uniq(cod_order_wcs)) as order_count,
                {% elif x.startswith('amount') %}
                        sumIf(amount * joinGet({{TABLE('stores_join', quote="'")}}, 'exchange_rate', toInt32(purchase_location))
                        ,
                            cod_transaction IN
                            {{ {'net': (72, 772, 73, 808), 'net_req': (72, 772, -73, -808), 'return': (73, 808), 'gross': (72, 772), 'return_req': (-73, -808) }[x.split('_', 1)[1]] }}
                        )
                {% elif x.startswith('units') %}
                        toInt64(sumIf(units, cod_transaction IN
                            {{ { 'net': (72, 772, 73, 808), 'net_req': (72, 772, -73, -808), 'return': (73, 808), 'gross': (72, 772), 'return_req': (-73, -808) }[x.split('_', 1)[1]] }}
                        ))
                {% end %}
                as {{symbol(x)}}
                {% if not last %},{% end %}
                {% end %}
                FROM (
                    WITH (select max(last_date) from sales_historic_last_date) as split_date
                    select cod_brand, sku_rank_lc, local_timeplaced, purchase_location, amount, cod_order_wcs, cod_transaction, units, cod_device, parent_catentry
                    {% if defined(commercial_tag) %}
                        ,
                        {% comment "this is a double join because the commercial tags filter" %}
                        joinGet({{TABLE('article_commercial_tags_dates_join', quote="'")}}, 'date_range',
                            concat(
                                joinGet({{TABLE('stores_join', quote="'")}}, 'country', toInt32(purchase_location)),
                                '_',
                                joinGet({{TABLE('articles_join', quote="'")}}, 'parent_partnumber', sku_rank_lc),
                                '_',
                                {{commercial_tag}}
                            )
                        ) as tag_range
                    {% end %}
                    from sales_rt
                    PREWHERE
                        local_timeplaced between toDateTime(toDate({{Date(date_start)}})) and toDateTime(toDate({{Date(date_end)}}) + 1) - 1
                    WHERE
                        local_timeplaced > toDateTime(split_date)
                        {% if defined(commercial_tag) %}
                            and local_timeplaced between toDateTime(tag_range[1]) and toDateTime(tag_range[2] + 1) - 1
                        {% end %}
                    and (cod_status <> 'X' or cod_status is null) and replacement = 0
                    {% comment "moved country filters here because there is an index by purchase_location" %}
                    {% if defined(country) or defined(cod_brand) %}
                        and purchase_location in (SELECT toInt16(location) from all_stores
                            where {{sql_and(country__in=Array(country, defined=False), cod_brand__in=Array(cod_brand, 'Int32', defined=False))}})
                    {% end %}
                    {% if defined(purchase_location) %}
                        and purchase_location in {{Array(purchase_location, 'Int32')}}
                    {% end %}
                    {% if defined(cod_order_type) %}
                        and cod_order_type = {{Int32(cod_order_type)}}
                    {% end %}
                    UNION ALL
                    WITH (select max(last_date) from sales_historic_last_date) as split_date
                    select cod_brand,CAST(sku_rank AS LowCardinality(String)) sku_rank_lc,local_timeplaced, purchase_location, amount, cod_order_wcs, cod_transaction, units, cod_device, parent_catentry
                    {% if defined(commercial_tag) %}
                        ,
                        {% comment "this is a double join because the commercial tags filter" %}
                        joinGet({{TABLE('article_commercial_tags_dates_join', quote="'")}}, 'date_range',
                            concat(
                                joinGet({{TABLE('stores_join', quote="'")}}, 'country', toInt32(purchase_location)),
                                '_',
                                joinGet({{TABLE('articles_join', quote="'")}}, 'parent_partnumber', sku_rank),
                                '_',
                                {{commercial_tag}}
                            )
                        ) as tag_range
                    {% end %}
                    from sales_historic_landing
                    WHERE
                        local_timeplaced between toDateTime(toDate({{Date(date_start)}})) and toDateTime(toDate({{Date(date_end)}}) + 1) - 1
                        and local_timeplaced <= toDateTime(split_date)
                        {% if defined(commercial_tag) %}
                            and local_timeplaced between toDateTime(tag_range[1]) and toDateTime(tag_range[2] + 1) - 1
                        {% end %}
                    and (cod_status <> 'X' or cod_status is null) and replacement = 0
                    {% comment "moved country filters here because there is an index by purchase_location" %}
                    {% if defined(country) or defined(cod_brand) %}
                        and purchase_location in (SELECT toInt16(location) from all_stores
                            where {{sql_and(country__in=Array(country, defined=False), cod_brand__in=Array(cod_brand, 'Int32', defined=False))}})
                    {% end %}
                    {% if defined(purchase_location) %}
                        and purchase_location in {{Array(purchase_location, 'Int32')}}
                    {% end %}
                    {% if defined(cod_order_type) %}
                        and cod_order_type = {{Int32(cod_order_type)}}
                    {% end %}
                )
                WHERE local_timeplaced + transform(purchase_location, arrayConcat([0], countries), arrayConcat([0], timezones)) < {% if end_ts_utc == 'now' %} now() {% else %} {{DateTime(end_ts_utc, '2019-01-01 00:00:00')}} {% end %}
                -- cod_device filter (optional)
                {% if defined(cod_device) %}
                    and cod_device in {{Array(cod_device)}}
                {% end %}
                {% if defined(cod_section) or defined(cod_product) or defined(cod_family) or defined(campaign) or defined(cod_product_line) or defined(cod_subfamily) or defined(partnumber) %}
                    and concat(
                        toString(cod_brand),
                        sku_rank_lc
                    ) in (
                        select concat(toString(cod_brand), sku_rank_lc) from {{TABLE('articles_filters')}}
                        prewhere 1
                        {% if defined(cod_section) or defined(cod_product) or defined(cod_family) or defined(campaign) or defined(cod_product_line) or defined(cod_subfamily) %}
                            AND {{sql_and(cod_section__in=Array(cod_section, 'Int',defined=False), cod_product__in=Array(cod_product,'Int', defined=False),cod_family__in=Array(cod_family, 'Int', defined=False), campaign__in=Array(campaign, defined=False), cod_product_line__in=Array(cod_product_line, 'Int', defined=False), cod_subfamily__in=Array(cod_subfamily, 'Int', defined=False))}}
                        {% end %}
                        {% if defined(country) or defined(cod_brand) %}
                            and cod_brand in (
                                SELECT cod_brand from all_stores
                                where {{sql_and(country__in=Array(country, defined=False), cod_brand__in=Array(cod_brand, 'Int32', defined=False))}}
                            )
                        {% end %}
                        {% if defined(partnumber) %}
                        AND
                            {% if not defined(campaign) %}
                                global_parent_partnumber
                            {% else %}
                                parent_partnumber
                            {% end %}
                            like {{partnumber}}
                        {% end %}
                    )
                {% end %}
                {% if defined(tags) %}
                and parent_catentry in (select parent_catentry from {{TABLE('tagging')}} where tag in {{Array(tags)}})
                {% end %}
                group by sku_rank_lc'''
            },
            {
                'name': 'historic',
                'sql': '''%
                    WITH (select max(last_date) from sales_historic_last_date) as split_date
                    SELECT
                        cast(sku_rank_lc as String) sku_rank_lc,
                        parent_catentry_array,
                        purchase_location,
                        {% if defined(commercial_tag) %}
                            {% comment "this is a double join because the commercial tags filter" %}
                            joinGet({{TABLE('article_commercial_tags_dates_join', quote="'")}}, 'date_range',
                                concat(
                                    joinGet({{TABLE('stores_join', quote="'")}}, 'country', toInt32(purchase_location)),
                                    '_',
                                    joinGet({{TABLE('articles_join', quote="'")}}, 'parent_partnumber', sku_rank_lc),
                                    '_',
                                    {{commercial_tag}}
                                )
                            ) as tag_range,
                        {% end %}
                        {% for last, x in enumerate_with_last(split_to_array(attr, 'amount_return')) %}
                        {% if not defined(cod_device) %}
                            {{symbol(x)}}
                            {% if x.startswith('amount') %}
                                    * joinGet({{TABLE('stores_join', quote="'")}}, 'exchange_rate', toInt32(purchase_location))
                            {% end %}
                        {% else %}
                            {% if x == 'order_count' %}
                            toInt32(arraySum(
                                arrayFilter(
                                    (order_wcs, device_hash) -> device_hash IN (select arrayJoin(arrayMap(x -> reinterpretAsUInt64(x), {{split_to_array(cod_device, '')}}))),
                                    `OrderMap.orders`,
                                    `OrderMap.cod_device_hash`
                                )
                            ))
                            {% else %}
                                arraySum(
                                    arrayFilter(
                                        (amount, cod_tx, device_hash) ->
                                            cod_tx IN {{ {'net': (72, 772, 73, 808), 'net_req': (72, 772, -73, -808), 'return': (73, 808), 'gross': (72, 772), 'return_req': (-73, -808) }[x.split('_', 1)[1]] }}
                                            and
                                            device_hash IN (select arrayJoin(arrayMap(x -> reinterpretAsUInt64(x), {{split_to_array(cod_device, '')}}))),
                                        {% if x.startswith('amount') %}
                                        `DeviceMap.amount`,
                                        {% else %}
                                        `DeviceMap.units`,
                                        {% end %}
                                        `DeviceMap.cod_transaction_Key`,
                                        `DeviceMap.cod_device_hash`
                                    )
                                )
                                {% if x.startswith('amount') %}
                                        * joinGet({{TABLE('stores_join', quote="'")}}, 'exchange_rate', toInt32(purchase_location))
                                {% end %}
                            {% end %}
                        {% end %}
                        as {{symbol(x)}}
                        {% if not last %},{% end %}
                        {% end %}
                    FROM
                    {% if defined(cod_order_type) %}
                        {{TABLE('product_rank_sint_mt')}}
                    {% else %}
                        {{TABLE('product_rank_mt')}}
                    {% end %}
                    WHERE
                        date between {{Date(date_start)}} and {{Date(date_end)}}
                        and date <= split_date
                        {% if not defined(commercial_tag) %}
                        and not date between
                            addMonths(
                                toStartOfMonth(toDate({{Date(date_start)}})),
                                if(toStartOfMonth(toDate({{Date(date_start)}})) = toDate({{Date(date_start)}}), 0, 1)
                            )
                            and
                            addMonths(
                                toStartOfMonth(toDate({{Date(date_end)}})),
                                if(toStartOfMonth(toDate({{Date(date_end)}})) = toStartOfMonth(toDate({{Date(date_end)}}) + 1), 0, 1)
                            ) - 1
                        {% else %}
                            and date between tag_range[1] and tag_range[2]
                        {% end %}
                    -- country filter
                    {% if defined(country) or defined(cod_brand) %}
                        and purchase_location in (SELECT toUInt16(location) from all_stores
                        where {{sql_and(country__in=Array(country, defined=False), cod_brand__in=Array(cod_brand, 'Int32', defined=False))}})
                    {% end %}
                    {% if defined(purchase_location) %}
                        and purchase_location in {{Array(purchase_location, 'Int32')}}
                    {% end %}'''
            },
            {
                'name': 'historic_month',
                'sql': '''%
                    WITH (select max(last_date) from sales_historic_last_date) as split_date
                    SELECT
                        sku_rank_lc,
                        parent_catentry_array,
                        purchase_location,
                        {% if defined(commercial_tag) %}
                            {% comment "this is a double join because the commercial tags filter" %}
                            joinGet({{TABLE('article_commercial_tags_dates_join', quote="'")}}, 'date_range',
                                concat(
                                    joinGet({{TABLE('stores_join', quote="'")}}, 'country', toInt32(purchase_location)),
                                    '_',
                                    joinGet({{TABLE('articles_join', quote="'")}}, 'parent_partnumber', sku_rank_lc),
                                    '_',
                                    {{commercial_tag}}
                                )
                            ) as tag_range,
                        {% end %}
                        {% for last, x in enumerate_with_last(split_to_array(attr, 'amount_return')) %}
                        {% if not defined(cod_device) %}
                            {{symbol(x)}}
                            {% if x.startswith('amount') %}
                                    * joinGet({{TABLE('stores_join', quote="'")}}, 'exchange_rate', toInt32(purchase_location))
                            {% end %}
                        {% else %}
                            {% if x == 'order_count' %}
                            toInt32(arraySum(
                                arrayFilter(
                                    (order_wcs, device_hash) -> device_hash IN (select arrayJoin(arrayMap(x -> reinterpretAsUInt64(x), {{split_to_array(cod_device, '')}}))),
                                    `OrderMap.orders`,
                                    `OrderMap.cod_device_hash`
                                )
                            ))
                            {% else %}
                                arraySum(
                                    arrayFilter(
                                        (amount, cod_tx, device_hash) ->
                                            cod_tx IN {{ {'net': (72, 772, 73, 808), 'net_req': (72, 772, -73, -808), 'return': (73, 808), 'gross': (72, 772), 'return_req': (-73, -808) }[x.split('_', 1)[1]] }}
                                            and
                                            device_hash IN (select arrayJoin(arrayMap(x -> reinterpretAsUInt64(x), {{split_to_array(cod_device, '')}}))),
                                        {% if x.startswith('amount') %}
                                        `DeviceMap.amount`,
                                        {% else %}
                                        `DeviceMap.units`,
                                        {% end %}
                                        `DeviceMap.cod_transaction_Key`,
                                        `DeviceMap.cod_device_hash`
                                    )
                                )
                                {% if x.startswith('amount') %}
                                        * joinGet({{TABLE('stores_join', quote="'")}}, 'exchange_rate', toInt32(purchase_location))
                                {% end %}
                            {% end %}
                        {% end %}
                        as {{symbol(x)}}
                        {% if not last %},{% end %}
                        {% end %}
                    FROM
                    {% if defined(cod_order_type) %}
                        {{TABLE('product_rank_sint_month_mt')}}
                    {% else %}
                        {{TABLE('product_rank_month_mt')}}
                    {% end %}
                    WHERE
                        date BETWEEN
                            addMonths(
                                toStartOfMonth(toDate({{Date(date_start)}})),
                                if(toStartOfMonth(toDate({{Date(date_start)}})) = toDate({{Date(date_start)}}), 0, 1)
                            )
                            and
                            addMonths(
                                toStartOfMonth(toDate({{Date(date_end)}})),
                                if(toStartOfMonth(toDate({{Date(date_end)}})) = toStartOfMonth(toDate({{Date(date_end)}}) + 1), 0, 1)
                            ) - 1
                            {% if defined(commercial_tag) %}
                                and date between tag_range[1] and tag_range[2]
                            {% end %}
                    -- country filter
                    {% if defined(country) or defined(cod_brand) %}
                        and purchase_location in (SELECT toUInt16(location) from all_stores
                        where {{sql_and(country__in=Array(country, defined=False), cod_brand__in=Array(cod_brand, 'Int32', defined=False))}})
                    {% end %}
                    {% if defined(purchase_location) %}
                        and purchase_location in {{Array(purchase_location, 'Int32')}}
                    {% end %}'''
            },
            {
                'name': 'rt',
                'sql': '''%
                    WITH (select max(last_date) from sales_historic_last_date) as split_date
                    SELECT
                        cast(sku_rank_lc as String) sku_rank_lc,
                        parent_catentry_array,
                        purchase_location,
                        {% if defined(commercial_tag) %}
                            {% comment "this is a double join because the commercial tags filter" %}
                            joinGet({{TABLE('article_commercial_tags_dates_join', quote="'")}}, 'date_range',
                                concat(
                                    joinGet({{TABLE('stores_join', quote="'")}}, 'country', toInt32(purchase_location)),
                                    '_',
                                    joinGet({{TABLE('articles_join', quote="'")}}, 'parent_partnumber', sku_rank_lc),
                                    '_',
                                    {{commercial_tag}}
                                )
                            ) as tag_range,
                        {% end %}
                        {% for last, x in enumerate_with_last(split_to_array(attr, 'amount_return')) %}
                        {% if not defined(cod_device) %}
                            {{symbol(x)}}
                            {% if x.startswith('amount') %}
                                    * joinGet({{TABLE('stores_join', quote="'")}}, 'exchange_rate', toInt32(purchase_location))
                            {% end %}
                        {% else %}
                            {% if x == 'order_count' %}
                            toInt32(arraySum(
                                arrayFilter(
                                    (order_wcs, device_hash) -> device_hash IN (select arrayJoin(arrayMap(x -> reinterpretAsUInt64(x), {{split_to_array(cod_device, '')}}))),
                                    `OrderMap.orders`,
                                    `OrderMap.cod_device_hash`
                                )
                            ))
                            {% else %}
                                arraySum(
                                    arrayFilter(
                                        (amount, cod_tx, device_hash) ->
                                            cod_tx IN {{ {'net': (72, 772, 73, 808), 'net_req': (72, 772, -73, -808), 'return': (73, 808), 'gross': (72, 772), 'return_req': (-73, -808) }[x.split('_', 1)[1]] }}
                                            and
                                            device_hash IN (select arrayJoin(arrayMap(x -> reinterpretAsUInt64(x), {{split_to_array(cod_device, '')}}))),
                                        {% if x.startswith('amount') %}
                                        `DeviceMap.amount`,
                                        {% else %}
                                        `DeviceMap.units`,
                                        {% end %}
                                        `DeviceMap.cod_transaction_Key`,
                                        `DeviceMap.cod_device_hash`
                                    )
                                )
                                {% if x.startswith('amount') %}
                                        * joinGet({{TABLE('stores_join', quote="'")}}, 'exchange_rate', toInt32(purchase_location))
                                {% end %}
                            {% end %}
                        {% end %}
                        as {{symbol(x)}}
                        {% if not last %},{% end %}
                        {% end %}
                    FROM
                    {% if defined(cod_order_type) %}
                        {{TABLE('product_rank_rt_sint_mt')}}
                    {% else %}
                        {{TABLE('product_rank_rt_mt')}}
                    {% end %}
                    WHERE date between {{Date(date_start)}} and {{Date(date_end)}} and date > split_date
                    {% if defined(commercial_tag) %}
                        and date between tag_range[1] and tag_range[2]
                    {% end %}
                    -- country filter
                    {% if defined(country) or defined(cod_brand) %}
                        and purchase_location in (SELECT toUInt16(location) from all_stores
                        where {{sql_and(country__in=Array(country, defined=False), cod_brand__in=Array(cod_brand, 'Int32', defined=False))}})
                    {% end %}
                    {% if defined(purchase_location) %}
                        and purchase_location in {{Array(purchase_location, 'Int32')}}
                    {% end %}'''
            },
            {
                'name': 'rank',
                'sql': '''%
                    SELECT
                        joinGet({{TABLE('articles_join', quote="'")}},
                            {% if not defined(campaign) %}
                                'global_parent_partnumber'
                            {% else %}
                                'parent_partnumber'
                            {% end %}, sku_rank_lc) product_rank,
                        groupUniqArray(sku_rank_lc) children,
                        groupUniqArrayMerge(parent_catentry_array) as parent_catentry_array,
                        {% for last, x in  enumerate_with_last(split_to_array(attr, 'amount_return')) %}
                            sum({{symbol(x)}}) as {{symbol(x)}}
                            {% if not last %},{% end %}
                        {% end %}
                    FROM
                    (
                        select * from (
                            {% if defined(end_ts_utc) %}
                                SELECT * FROM by_timestamp
                            {% else %}
                                SELECT * from historic
                                UNION ALL
                                SELECT * from rt
                                {% if not defined(commercial_tag) %}
                                UNION ALL
                                SELECT * from historic_month
                                {% end %}
                            {% end %}
                        )
                        {% if not defined(end_ts_utc) %}
                        WHERE 1
                        {% if defined(cod_section) or defined(cod_product) or defined(cod_family) or defined(campaign) or defined(cod_product_line) or defined(cod_subfamily) or defined(partnumber) %}
                            and concat(
                                toString(joinGet({{TABLE('stores_join', quote="'")}}, 'cod_brand', toInt32(purchase_location))),
                                sku_rank_lc
                            ) in (
                                select concat(toString(cod_brand), sku_rank_lc) from {{TABLE('articles_filters')}}
                                prewhere 1
                                {% if defined(cod_section) or defined(cod_product) or defined(cod_family) or defined(campaign) or defined(cod_product_line) or defined(cod_subfamily) %}
                                    AND {{sql_and(cod_section__in=Array(cod_section, 'Int',defined=False), cod_product__in=Array(cod_product,'Int', defined=False),cod_family__in=Array(cod_family, 'Int', defined=False), campaign__in=Array(campaign, defined=False), cod_product_line__in=Array(cod_product_line, 'Int', defined=False), cod_subfamily__in=Array(cod_subfamily, 'Int', defined=False))}}
                                {% end %}
                                {% if defined(country) or defined(cod_brand) %}
                                    and cod_brand in (
                                        SELECT distinct cod_brand from all_stores
                                        where {{sql_and(country__in=Array(country, defined=False), cod_brand__in=Array(cod_brand, 'Int32', defined=False))}}
                                    )
                                {% end %}
                                {% if defined(partnumber) %}
                                AND
                                    {% if not defined(campaign) %}
                                        global_parent_partnumber
                                    {% else %}
                                        parent_partnumber
                                    {% end %}
                                    like {{partnumber}}
                                {% end %}
                            )
                        {% end %}
                        {% end %}
                    )
                    WHERE product_rank != ''
                    GROUP BY product_rank
                    {% if defined(tags) %}
                    having arraySum(arrayMap(x -> toInt64(x) in (select parent_catentry from {{TABLE('tagging')}} where tag in {{Array(tags)}}) ? 1: 0,
                        parent_catentry_array
                    )) > 0
                    {% end %}
                    ORDER BY {{symbol(split_to_array(attr, 'amount_return')[0])}}
                        {% if defined(sort) and sort.lower() == 'asc' %}
                        ASC
                        {% else %}
                        DESC
                        {% end %}
                    LIMIT {{Int32(page_size, 32)}}
                    OFFSET {{Int32(page, 0) * Int32(page_size, 32)}}'''
            },
            {
                'name': 'sales_in_units_and_amount_by_country',
                'sql': '''%
                {% if defined(commercial_tag) %}
                    {{max_threads(10)}}
                {% else %}
                    {{max_threads(8)}}
                {% end %}
                SELECT
                        product_rank as sku,
                        {% if not defined(campaign) %}
                        joinGet({{TABLE('articles_join_global_partnumber', quote="'")}}, 'image', product_rank) as image,
                        joinGet({{TABLE('articles_join_global_partnumber', quote="'")}}, 'plain_image', product_rank) as plain_image,
                        joinGet({{TABLE('articles_join_global_partnumber', quote="'")}}, 'description', product_rank) as description
                        {% else %}
                        joinGet({{TABLE('articles_join_parent_partnumber', quote="'")}}, 'image', product_rank) as image,
                        joinGet({{TABLE('articles_join_parent_partnumber', quote="'")}}, 'plain_image', product_rank) as plain_image,
                        joinGet({{TABLE('articles_join_parent_partnumber', quote="'")}}, 'description', product_rank) as description
                    {% end %},
                    arrayMap(
                        x -> joinGet({{TABLE('articles_join', quote="'")}}, 'parent_partnumber', x),
                        children
                    ) as partnumbers,
                    arrayDistinct(
                    arrayFilter(
                        x -> x != '',
                        arrayFlatten(
                            arrayMap(x-> joinGet('tagging_array', 'tags', toInt64(x)), parent_catentry_array)
                        )
                    )
                    ) as tags,
                    arrayDistinct(arrayFilter(
                        x ->
                            -- {tag} -> x.1, {country} -> x.2, {start_date} -> x.3, {end_date} -> x.4
                            -- this is the basic algorithm to check if 1d segments overlap: https://eli.thegreenplace.net/2008/08/15/intersection-of-1d-segments
                            {{Date(date_end)}} >= x.3 and  x.4 >= {{Date(date_start)}}
                            {% if defined(country) %} and x.2 in {{Array(country)}} {% end %}
                            {% if defined(commercial_tag) %} and x.1 in {{Array(commercial_tag)}} {% end %},
                    arrayMap(x -> arrayMap(country, start_date, end_date, tag -> (tag, country, start_date, end_date), x.1, x.2, x.3, x.4) ,
                        arrayMap(partnumber ->
                        {% if not defined(campaign) %}
                        joinGet({{TABLE('article_commercial_tags_join_global_parent_partnumber', quote="'")}}, 'groupped_fields', partnumber),
                        {% else %}
                        joinGet({{TABLE('article_commercial_tags_join_parent_partnumber', quote="'")}}, 'groupped_fields', partnumber),
                        {% end %}
                        partnumbers
                        )
                    )[1]
                    )) commercial_tags,
                    {% for last, x in  enumerate_with_last(split_to_array(attr, 'amount_return')) %}
                        {{symbol(x)}}
                        {% if not last %},{% end %}
                    {% end %}
                    FROM rank
                    {% if defined(symbol) %}
                        WHERE sku == {{String(symbol)}}
                    {% end %}
                    '''
            }]

        pipe_name = 'example'
        token_name = 'example endpoint api'
        u = Users.get_by_id(self.WORKSPACE_ID)
        with User.transaction(u.id) as user:
            for ds_name in datasources:
                user.add_datasource(ds_name)
            pipe = user.add_pipe(pipe_name, nodes=nodes)
            pipe.endpoint = pipe.pipeline.nodes[-1].id
            user.update_pipe(pipe)
            user.add_token(token_name, scopes.PIPES_READ, pipe.id)

        self.u = Users.get_by_id(self.WORKSPACE_ID)
        self.variables = {
            'date_start': '2020-09-03',
            'date_end': '2020-09-03',
            'attr': 'order_count',
            'cod_section': '1',
            'cod_brand': '1,16',
            'page': '0',
            'page_size': '200',
        }

        self.pipe = self.u.get_pipe(pipe_name)
        self.template_execution_results = {}
        access_info = self.u.get_token_access_info(token_name)
        self.readable_resources = access_info.get_resources_for_scope(scopes.DATASOURCES_READ, scopes.PIPES_READ)

        self.assertEqual(User.replace_executor._max_workers, 1, "This test is designed for 1 worker only")

        await self.replace_tables_with_options(flush=False, randomize=None, sync=False)  # Ensure warmup (for processes)
        await IOLoop.current().run_in_executor(User.replace_executor, flush_all_caches)
        timings_with_cache = await self.get_replace_timings(flush=False, randomize='cod_section')

        hits, misses = await IOLoop.current().run_in_executor(User.replace_executor, get_cache_info)
        self.assertEqual(misses, 5)
        self.assertEqual(hits, 5 * (self.ITERATIONS - 1))

        total_cached = sum(timings_with_cache)
        print(f"Total with cache: {total_cached}")
        self.check_timings(timings_with_cache, max_mean=.08, max_p90=.09, max_min=.07)

        timings_without_cache = await self.get_replace_timings(flush=True, randomize='cod_section')

        total_without_cache = sum(timings_without_cache)
        print(f"Total without cache: {total_without_cache}")
        self.check_timings(timings_without_cache, max_mean=.16, max_p90=.18, max_min=.14)

        improvement = total_without_cache / total_cached
        print(f"Improvement: {improvement:.3f}x")
        self.assertGreater(improvement, 1.5, "Timings with cache should be much faster than without it")

    @tornado.testing.gen_test(timeout=360)
    async def test_toolset_cache_constant_parameter(self):

        datasources = [
            "prd_rawdata.ds_stm_wpv",
            "prd_benchdirect.currencies_exchange_join",
            "prd_rawdata.ds_mv_currencies_join__v1",
            "prd_rawdata.ds_org_currencies_historical_join",
            "fingerprint_property_total_origins",
            "fingerprint_property_group_by_day",
            "fingerprint_group_by_day",
            "property_grouped",
            "fingerprint_grouped",
            "fingerprint_chain_grouped",
            "prd_rawdata.ds_stm_events__v1",
            "fingerprint_property_group_by_tool",
            "fingerprint_chain_group_by_tool",
            "prd_rawdata.ds_mv_partner_hotels_join_by_property_id__v2",
            "prd_rawdata.ds_mv_pages_join_by_id__v2",
            "fingerprint_property_group_by_toolgroup",
            "prd_rawdata.ds_db_toolgroup_tools",
            "fingerprint_corporative_grouped",
            "fingerprint_property_grouped",
            "fingerprint_corporative_group_by_day",
            "fingerprint_corporative_group_by_tool",
            "fingerprint_corporative_group_by_toolgroup",
            "ds_test",
            "booking_ids",
            "fingerprint_chain_group_by_day",
            "browser_version_group_by_day",
            "browser_group_by_day",
            "country_group_by_day",
            "device_group_by_day",
            "geoname_group_by_day",
            "language_group_by_day",
            "network_group_by_day",
            "os_group_by_day",
            "os_version_group_by_day",
            "property_group_by_day",
        ]

        nodes = [
            {
                "name": "fingerprint_property_node",
                "sql": '''%
                    SELECT
                        -- Visits
                        countMerge(review_visits_shown) as user_property_total_reviews_shown,
                        -- Searches
                        countMerge(saved_search_visits) as user_property_total_saved_searches,
                        -- Package/Price
                        countMerge(total_prices) as user_property_total_prices,
                        max(last_price_date) as user_property_last_price_date,
                        argMaxMerge(last_search_type) as user_property_last_search_type,
                        argMaxMerge(last_search_price) as user_property_last_search_price,
                        argMaxMerge(last_search_currency) as user_property_last_search_currency,
                        argMaxMerge(last_search_start_date) as user_property_last_search_start_date,
                        argMaxMerge(last_search_end_date) as user_property_last_search_end_date,
                        argMaxMerge(last_search_adults) as user_property_last_search_adults,
                        argMaxMerge(last_search_children) as user_property_last_search_children,
                        argMaxMerge(last_search_babies) as user_property_last_search_babies,
                        argMaxMerge(last_search_rooms) as user_property_last_search_rooms,
                        argMaxMerge(last_search_status) as user_property_last_search_status,
                        argMaxMerge(last_search_ab_test) as user_property_last_search_ab_test,
                        argMaxMerge(last_search_showed) as user_property_last_search_showed,
                        -- Last visit
                        max(last_visit_date) as user_property_last_visit_date,
                        countMerge(total_visits) as user_property_total_visits,
                        uniqMerge(total_visit_days) as user_property_total_visit_days,
                        max(last_index_date) as user_property_last_index_date,
                        argMaxMerge(last_visit_page_id) as user_property_last_visit_page_id,
                        argMaxMerge(last_visit_page_name) as user_property_last_visit_page_name,
                        argMaxMerge(last_funnel_page_name) as user_property_last_funnel_page_name,
                        argMaxMergeIf(furthest_funnel_page_name, datediff('day', (furthest_funnel_date), now()) <= 30) as user_property_furthest_funnel_page_name,
                        argMaxMerge(last_visit_device) as user_property_last_device,
                        argMaxMerge(last_search_url) as user_property_last_search_url,
                        argMaxMerge(last_visit_referrer) as user_property_last_visit_referrer,
                        -- Bookings
                        countMerge(booking_visits) as user_property_total_bookings,
                        max(last_booking_date) as user_property_last_booking_date,
                        -- Origins
                        max(last_origin_date) as user_property_last_origin_date,
                        toDateTime(topKWeightedMerge(2)(last_origin_dates_uint32)[2]) as user_property_second_to_last_origin_date,
                        argMaxMerge(last_origin_source) as user_property_last_origin_source,
                        argMaxMerge(last_origin_retargeting) as user_property_last_origin_retargeting,
                        argMaxMerge(last_origin_browser) as user_property_last_origin_browser,
                        argMaxMerge(last_origin_browser_major) as user_property_last_origin_browser_major,
                        argMaxMerge(last_origin_browser_version) as user_property_last_origin_browser_version,
                        argMaxMerge(last_origin_os) as user_property_last_origin_os,
                        argMaxMerge(last_origin_os_version) as user_property_last_origin_os_version,
                        argMaxMerge(last_origin_user_agent) as user_property_last_origin_user_agent,
                        argMaxMerge(last_origin_ntw_booked) as user_property_last_origin_ntw_booked,
                        argMaxMerge(last_origin_ntw_visits) as user_property_last_origin_ntw_visits,
                        argMaxMerge(last_origin_ntw_chains) as user_property_last_origin_ntw_chains,
                        argMaxMerge(last_visit_geoname_id) as user_property_last_visit_geoname_id,
                        countMerge(total_origin_search) as user_property_total_origin_search,
                        countMerge(total_origin_facebook) as user_property_total_origin_facebook,
                        countMerge(total_origin_twitter) as user_property_total_origin_twitter,
                        countMerge(total_origin_instagram) as user_property_total_origin_instagram,
                        countMerge(total_origin_kayak) as user_property_total_origin_kayak,
                        countMerge(total_origin_tripadvisor) as user_property_total_origin_tripadvisor,
                        countMerge(total_origin_trivago) as user_property_total_origin_trivago,
                        countMerge(total_origin_direct) as user_property_total_origin_direct,
                        countMerge(total_origin_noorigin) as user_property_total_origin_noorigin,
                        countMerge(total_origin_retargeting) as user_property_total_origin_retargeting,
                        -- tools
                        countMerge(exit_visits_shown) as user_property_total_exits_shown,
                        countMerge(inline_visits_shown) as user_property_total_inlines_shown,
                        countMerge(layer_visits_shown) as user_property_total_layers_shown,
                        countMerge(note_visits_shown) as user_property_total_notes_shown
                    FROM fingerprint_property_grouped c
                    PREWHERE cityHash64({{String(fingerprint, 'aa96f1b6362a46682193c4708cedfaff', required=True)}}, lower({{String(country, 'es', required=True)}}), lower({{String(device, 'desktop', required=True)}}), toUInt128({{UInt128(property_id, '1027947')}})) == c.key'''
            },
            {
                "name": "results",
                "sql": '''%
                    {{max_threads(1)}}
                    SELECT
                        cityHash64({{String(fingerprint, 'aa96f1b6362a46682193c4708cedfaff', required=True)}}, lower({{String(country, 'es', required=True)}}), lower({{String(device, 'desktop', required=True)}}), toUInt128({{UInt128(property_id, '1027947')}})) as key,
                        -- fingerprint_property_node
                        if(user_property_last_origin_date > '1970-01-01 00:00:00', user_property_last_origin_date, null) as user_property_last_origin_date,
                        if(user_property_second_to_last_origin_date > '1970-01-01 00:00:00', user_property_second_to_last_origin_date, null) as user_property_second_to_last_origin_date,
                        user_property_last_origin_source,
                        user_property_last_origin_retargeting,
                        user_property_last_origin_browser,
                        user_property_last_origin_browser_major,
                        user_property_last_origin_browser_version,
                        user_property_last_origin_os,
                        user_property_last_origin_os_version,
                        user_property_last_origin_user_agent,
                        user_property_last_origin_ntw_booked,
                        user_property_last_origin_ntw_visits,
                        user_property_last_origin_ntw_chains,
                        user_property_last_visit_geoname_id,
                        user_property_total_origin_search,
                        user_property_total_origin_facebook,
                        user_property_total_origin_twitter,
                        user_property_total_origin_instagram,
                        user_property_total_origin_kayak,
                        user_property_total_origin_tripadvisor,
                        user_property_total_origin_trivago,
                        user_property_total_origin_direct,
                        user_property_total_origin_noorigin,
                        user_property_total_origin_retargeting,
                        user_property_total_bookings,
                        if(user_property_last_booking_date > '1970-01-01 00:00:00', user_property_last_booking_date, null) as user_property_last_booking_date,
                        user_property_total_exits_shown,
                        user_property_total_inlines_shown,
                        user_property_total_layers_shown,
                        user_property_total_notes_shown,
                        CASE
                            WHEN (user_property_last_origin_source <> 'google' and user_property_last_origin_source <> 'facebook' and user_property_last_origin_source <> 'twitter' and user_property_last_origin_source <> 'instagram' and user_property_last_origin_source <> 'kayak' and user_property_last_origin_source <> 'tripadvisor' and user_property_last_origin_source <> 'trivago' and user_property_last_origin_source <> '') THEN 'direct'
                            WHEN user_property_last_origin_source = 'google' THEN 'search'
                            WHEN user_property_last_origin_source = '' THEN 'noorigin'
                            ELSE user_property_last_origin_source
                        END as user_property_last_origin,
                        user_property_total_reviews_shown,
                        user_property_total_saved_searches,
                        user_property_total_prices,
                        if(user_property_last_price_date > '1970-01-01 00:00:00', user_property_last_price_date, null) as user_property_last_price_date,
                        user_property_last_search_type,
                        user_property_last_search_price,
                        user_property_last_search_currency,
                        if(user_property_last_search_start_date > '1970-01-01', user_property_last_search_start_date, null) as user_property_last_search_start_date,
                        if(user_property_last_search_end_date > '1970-01-01', user_property_last_search_end_date, null) as user_property_last_search_end_date,
                        user_property_last_search_adults,
                        user_property_last_search_children,
                        user_property_last_search_babies,
                        user_property_last_search_rooms,
                        user_property_last_search_status,
                        user_property_last_search_ab_test,
                        user_property_last_search_showed,
                        if(user_property_last_visit_date > '1970-01-01 00:00:00', user_property_last_visit_date, null) as user_property_last_visit_date,
                        user_property_total_visits,
                        user_property_total_visit_days,
                        if(user_property_last_index_date > '1970-01-01 00:00:00', user_property_last_index_date, null) as user_property_last_index_date,
                        user_property_last_visit_page_id,
                        user_property_last_visit_page_name,
                        user_property_last_funnel_page_name,
                        user_property_last_device,
                        user_property_last_search_url,
                        user_property_last_visit_referrer,
                        CASE user_property_last_funnel_page_name
                            WHEN 'Calendar' THEN 2
                            WHEN 'Rooms and Rates' THEN 3
                            WHEN 'Packages' THEN 3
                            WHEN 'User Register Page' THEN 4
                            WHEN 'Booking Confirmed' THEN 5
                            ELSE 0
                        END as user_property_last_funnel_page_code
                    FROM
                        fingerprint_property_node''',
            }
        ]

        pipe_name = 'example'
        token_name = 'example endpoint api'
        u = Users.get_by_id(self.WORKSPACE_ID)
        with User.transaction(u.id) as user:
            for ds_name in datasources:
                user.add_datasource(ds_name)
            pipe = user.add_pipe(pipe_name, nodes=nodes)
            pipe.endpoint = pipe.pipeline.nodes[-1].id
            user.update_pipe(pipe)
            user.add_token(token_name, scopes.PIPES_READ, pipe.id)

        self.u = Users.get_by_id(self.WORKSPACE_ID)
        self.variables = {
            'fingerprint': 'anything'
        }

        self.pipe = self.u.get_pipe(pipe_name)
        self.template_execution_results = {}
        access_info = self.u.get_token_access_info(token_name)
        self.readable_resources = access_info.get_resources_for_scope(scopes.DATASOURCES_READ, scopes.PIPES_READ)

        self.assertEqual(User.replace_executor._max_workers, 1, "This test is designed for 1 worker only")

        await self.replace_tables_with_options(flush=False, randomize=None, sync=False)  # Ensure warmup (for processes)
        await IOLoop.current().run_in_executor(User.replace_executor, get_template_and_variables.cache_clear)
        timings_with_cache = await self.get_replace_timings(flush=False, randomize=None)

        total_cached = sum(timings_with_cache)
        print(f"Total with cache: {total_cached}")
        self.check_timings(timings_with_cache, max_mean=.04, max_p90=.045, max_min=.035)

        timings_without_cache = await self.get_replace_timings(flush=True, randomize=None)

        total_without_cache = sum(timings_without_cache)
        print(f"Total without cache: {total_without_cache}")
        self.check_timings(timings_without_cache, max_mean=.08, max_p90=.9, max_min=.7)

        improvement = total_without_cache / total_cached
        print(f"Improvement: {improvement:.3f}x")
        self.assertGreater(improvement, 1.3, "Timings with cache should be much faster than without it")



import asyncio
import unittest
import uuid
from typing import Optional
from unittest.mock import patch

import pytest
import tornado.testing
from tornado.testing import AsyncTestCase
from tests.conftest import CH_ADDRESS, ClusterPatches

from tinybird.app import uri_is_interna_or_from_api
from tinybird.constants import CHCluster
from tinybird.job import JobExecutor
from tinybird.user import Users, User, public, UserAccount
from tinybird.internal_resources import init_metrics_tables, init_internal_tables
from tinybird.ch import ch_table_schema, ch_drop_table_sync, HTTPClient, ch_drop_view, ch_source_table_for_view_sync
from tinybird.ch_utils.exceptions import CHException
from tinybird.default_tables import DEFAULT_METRICS_TABLES, DEFAULT_METRICS_VIEWS, DEFAULT_TABLES, DefaultTable, \
    DEFAULT_METRICS_CLUSTER_TABLES, DEFAULT_METRICS_CLUSTER_VIEWS, DEFAULT_VIEWS
from tinybird.redis_client import get_redis_args, get_redis_client
from tinybird.syncasync import async_to_sync
from tinybird.constants import BillingPlans


class BaseTestInitTables(AsyncTestCase):

    @classmethod
    def setUpClass(cls):
        cls.original_public_user = public.get_public_email()
        cls.original_public_database = public.get_public_database()

        cls.workspace_name = f'r_test_internal_{uuid.uuid4().hex}'
        cls.user_email = f'{cls.workspace_name}@localhost'

        cls.user_account = UserAccount.register(cls.user_email, 'pass')
        cls.workspace = User.register(cls.workspace_name, cls.user_account.id, CHCluster(name='tinybird', server_url=CH_ADDRESS))

        client = HTTPClient(cls.workspace.database_server, database=None)
        client.query_sync(f"CREATE DATABASE IF NOT EXISTS `{cls.workspace.database}` ON CLUSTER tinybird", read_only=False)
        public.set_public_user(cls.user_email, cls.workspace.database)
        redis_client = get_redis_client()
        cls.job_executor = JobExecutor(redis_client=redis_client, redis_kwargs=get_redis_args(), consumer=True, import_workers=1, query_workers=1)

    @classmethod
    def tearDownClass(cls) -> None:
        database = cls.workspace.database
        client = HTTPClient(cls.workspace.database_server, database=None)
        client.query_sync(f"DROP DATABASE IF EXISTS `{database}` ON CLUSTER tinybird", read_only=False)
        User._delete(cls.workspace.id)
        UserAccount._delete(cls.user_account.id)
        public.set_public_user(cls.original_public_user, cls.original_public_database)

    def setUp(self):
        super().setUp()
        public.set_public_user(self.user_email, self.workspace.database)
        self.created_tables = []
        self.created_views = []

    def tearDown(self):
        pu = self.workspace
        for table in self.created_tables:
            ds = Users.get_datasource(pu, table)
            if ds:
                ch_drop_table_sync(pu.database_server, pu.database, ds.id, exists_clause=True)
                Users.drop_datasource(pu, ds.name)
        for view in self.created_views:
            pipe = Users.get_pipe(pu, view)
            if pipe:
                ch_drop_view_sync = async_to_sync(ch_drop_view)
                ch_drop_view_sync(pu.database_server, pu.database, pipe.name)
                Users.drop_pipe(pu, pipe.name)
        self.job_executor._clean()
        super().tearDown()

    async def _init_tables(self, tables, views=None):
        if views is None:
            views = []
        await init_internal_tables(tables, views, populate_views=False, job_executor=self.job_executor)
        self.created_tables += [t.name for t in tables]
        self.created_views += [v.name for v in views]


class TestInitInternalTables(BaseTestInitTables):

    @tornado.testing.gen_test
    async def test_happy_case(self):
        await self._init_tables(DEFAULT_TABLES, DEFAULT_VIEWS)
        pu = public.get_public_user()
        for t in DEFAULT_TABLES:
            ds = pu.get_datasource(t.name)
            self.assertEqual(ds.name, t.name)
            self.assertEqual(ds.tags['__version'], len(t.migrations))
            schema = ch_table_schema(ds.id, pu.database_server, pu.database)
            self.assertIsNotNone(schema)
        for v in DEFAULT_VIEWS:
            view = pu.get_pipe(v.name)
            self.assertEqual(view.name, v.name)

            if v.datasource_name:
                self.assertEqual(Users.get_datasource(pu, v.datasource_name).id, view.pipeline.nodes[0].materialized)
                self.assertEqual(pu.cluster, view.pipeline.nodes[0].cluster)
                self.assertEqual(v.query, view.pipeline.nodes[0].sql)
                source_table = ch_source_table_for_view_sync(pu.database_server, pu.database, view.pipeline.nodes[0].id)
                self.assertIsNotNone(source_table)


class TestInitInternalTablesMigration(BaseTestInitTables):
    @tornado.testing.gen_test
    async def test_migration_happy_case(self):
        pu = public.get_public_user()
        name = 'foo_table'
        schema_at0 = 'timestamp DateTime, event String'
        engine = 'MergeTree() PARTITION BY toYear(timestamp) ORDER BY (event, timestamp)'
        t_at0 = DefaultTable(name, schema_at0, engine)
        await self._init_tables([t_at0])

        ds = Users.get_datasource(pu, t_at0.name)
        schema = ch_table_schema(ds.id, pu.database_server, pu.database)
        self.assertEqual(len(schema), 2)
        self.assertEqual(schema[0]['name'], 'timestamp')
        self.assertEqual(schema[1]['name'], 'event')

        schema_at1 = 'timestamp DateTime, event String, error Nullable(String)'
        migration = 'ADD COLUMN IF NOT EXISTS error Nullable(String) AFTER timestamp'
        t_at1 = DefaultTable(name, schema_at1, engine, [[migration]])
        await self._init_tables([t_at1])

        schema = ch_table_schema(ds.id, pu.database_server, pu.database)
        self.assertEqual(len(schema), 3)
        self.assertEqual(schema[0]['name'], 'timestamp')
        self.assertEqual(schema[1]['name'], 'error')
        self.assertTrue(schema[1]['nullable'])
        self.assertEqual(schema[2]['name'], 'event')


class TestInitInternalTablesFailure(BaseTestInitTables):
    @tornado.testing.gen_test
    async def test_migration_failure(self):
        pu = public.get_public_user()
        name = 'bar_table'
        schema_at0 = 'timestamp DateTime, event String'
        engine = 'MergeTree() PARTITION BY toYear(timestamp) ORDER BY (event, timestamp)'
        t_at0 = DefaultTable(name, schema_at0, engine)
        await self._init_tables([t_at0])

        schema_at1 = 'timestamp DateTime, event Nullable(String)'
        migration = 'MODIFY COLUMN IF EXISTS event Nullable(String)'
        t_at1 = DefaultTable(name, schema_at1, engine, [[migration]])
        # trying to make nullable a column in the table index

        with self.assertRaises(CHException) as exception:
            await self._init_tables([t_at1])
            message = str(exception)
            old_pattern = r'ALTER of key column event.*'
            new_pattern = r'Sorting key cannot contain nullable columns.*'
            self.assertTrue(old_pattern.match(message) or new_pattern.match(message))

        # table still exists and works as initial version
        ds = Users.get_datasource(pu, t_at0.name)
        self.assertEqual(ds.tags['__version'], 0)
        schema = ch_table_schema(ds.id, pu.database_server, pu.database)
        self.assertEqual(len(schema), 2)
        self.assertEqual(schema[0]['name'], 'timestamp')
        self.assertEqual(schema[1]['name'], 'event')
        self.assertFalse(schema[1]['nullable'])


class TestInitInternalTablesMigrationsV1(BaseTestInitTables):

    def _test_expectations(self, expectations):
        pu = public.get_public_user()
        for ds_name, version, n_columns, columns in expectations:
            ds = Users.get_datasource(pu, ds_name)
            self.assertEqual(ds.tags['__version'], version)
            schema = ch_table_schema(ds.id, pu.database_server, pu.database)
            self.assertEqual(len(schema), n_columns)
            for c_index, c_name, *attrs in columns:
                self.assertEqual(schema[c_index]['name'], c_name)
                for k, v in attrs:
                    self.assertEqual(schema[c_index][k], v)

    @tornado.testing.gen_test
    async def test_migration_happy_case(self):
        tables_v0 = [
            DefaultTable('js_errors', """timestamp DateTime,
                error_message String,
                error_type String,
                stackframes Nullable(String),
                browser_name String,
                browser_version String,
                os_name String,
                os_version String,
                user_email String,
                url String,
                steps Nullable(String)"""),
            DefaultTable('block_log', """timestamp DateTime,
                import_id String,
                source String,
                token_id String,
                block_id String,
                status String,
                user_id String,
                user_mail String,
                datasource_id String,
                datasource_name String,
                start_offset Nullable(Int64),
                end_offset Nullable(Int64),
                rows Nullable(Int32),
                parser Nullable(String),
                quarantine_lines Nullable(UInt32),
                empty_lines Nullable(UInt32),
                bytes Nullable(UInt32),
                processing_time Nullable(Float32),
                processing_error Nullable(String)"""),
            DefaultTable('hook_log', """timestamp DateTime,
                import_id String,
                source String,
                hook_id String,
                name String,
                operation String,
                status String,
                user_id String,
                user_mail String,
                datasource_id String,
                datasource_name String,
                processing_time Nullable(Float32),
                processing_error Nullable(String)"""),
            DefaultTable('datasources_ops_log', """timestamp DateTime,
                event_type String,
                datasource_id String,
                datasource_name String,
                user_id String,
                user_mail String,
                result String,
                elapsed_time Float32,
                error Nullable(String),
                job_id Nullable(String),
                rows Nullable(UInt64),
                rows_quarantine Nullable(UInt64),
                blocks_ids Array(String),
                Options Nested(
                    Names String,
                    Values String
                )""")
        ]

        await self._init_tables(tables_v0)

        self._test_expectations([
            ('js_errors', 0, 11, [
                (0, 'timestamp'),
                (-1, 'steps'),
            ]),
            ('block_log', 0, 19, [
                (0, 'timestamp'),
                (1, 'import_id'),
                (2, 'source'),
                (-1, 'processing_error'),
            ]),
            ('hook_log', 0, 13, [
                (0, 'timestamp'),
                (1, 'import_id', ('nullable', False)),
                (2, 'source'),
                (-1, 'processing_error'),
            ]),
            ('datasources_ops_log', 0, 15, [
                (0, 'timestamp'),
                (8, 'error', ('nullable', True)),
                (9, 'job_id', ('nullable', True)),
                (-1, 'Options.Values'),
            ]),
        ])

        tables_v1 = []
        for t in DEFAULT_TABLES:
            if t.name in [t.name for t in tables_v0]:
                migrations = []
                if t.migrations:
                    migrations.append(t.migrations[0])
                tables_v1.append(DefaultTable(t.name, t.schema, engine=t.engine, migrations=migrations))

        await self._init_tables(tables_v1)

        self._test_expectations([
            ('js_errors', 0, 11, [
                (0, 'timestamp'),
                (-1, 'steps'),
            ]),
            ('block_log', 1, 21, [
                (0, 'timestamp'),
                (1, 'request_id', ('nullable', False)),
                (2, 'import_id'),
                (3, 'job_id', ('nullable', True)),
                (4, 'source'),
                (-1, 'processing_error'),
            ]),
            ('hook_log', 1, 15, [
                (0, 'timestamp'),
                (1, 'request_id', ('nullable', False)),
                (2, 'import_id', ('nullable', True)),
                (3, 'job_id', ('nullable', True)),
                (4, 'source'),
                (-1, 'processing_error')
            ]),
            ('datasources_ops_log', 1, 17, [
                (0, 'timestamp'),
                (8, 'error', ('nullable', True)),
                (9, 'request_id', ('nullable', False)),
                (10, 'import_id', ('nullable', True)),
                (11, 'job_id', ('nullable', True)),
                (-1, 'Options.Values'),
            ]),
        ])


class TestInitMetricsTables(BaseTestInitTables):
    @patch.object(HTTPClient, 'query')
    @tornado.testing.gen_test
    async def test_query_templates_for_init_metrics_tables_for_internal(self, query_mock):
        empty_data_future = asyncio.Future()
        empty_data_future.set_result((None, '{"data": []}'))
        cluster_future = asyncio.Future()
        cluster_future.set_result((None,
                                   '{"data": [{"cluster": "cluster_1", "host_address": "127.0.0.1"},{"cluster": "cluster_2", "host_address": "127.0.0.2"}]}'))

        query_mock.side_effect = [
            empty_data_future,
            cluster_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
        ]

        public_user = public.get_public_user()
        public_user.clusters = ['internal']
        public_user.save()

        with patch.object(Users, 'add_datasource') as mock_add_datasource:
            await init_metrics_tables(
                host=public.get_public_user().database_server,
                metrics_cluster='metrics',
                metrics_database_server='http://127.0.0.1',
                metrics_cluster_tables=DEFAULT_METRICS_CLUSTER_TABLES,
                metrics_cluster_views=DEFAULT_METRICS_CLUSTER_VIEWS,
                metrics_tables=DEFAULT_METRICS_TABLES,
                metrics_views=DEFAULT_METRICS_VIEWS,
                metrics_database='default',
                add_datasources=True
            )

        self.assertEqual(mock_add_datasource.call_count, len(DEFAULT_METRICS_TABLES))

        queries = [' '.join(call.args[0].split()).replace(' ', '') for call in query_mock.call_args_list]

        self.assertTrue("CREATE DATABASE IF NOT EXISTS default on cluster metrics".replace(' ', '') in queries)
        self.assertTrue(
            """
            CREATE TABLE IF NOT EXISTS default.usage_metrics_log ON CLUSTER metrics(
                `date` Date,
                `database` String,
                `host` String,
                `read_bytes` UInt64,
                `written_bytes` UInt64
            )
            ENGINE = ReplicatedSummingMergeTree('/clickhouse/tables/{layer}-{shard}/default.usage_metrics_log', '{replica}') PARTITION BY toYYYYMM(date) ORDER BY (host, database, date)
            """.replace('\n', '').replace(' ', '') in queries, queries)

        self.assertTrue(f"CREATE DATABASE IF NOT EXISTS {public_user.database} ON CLUSTER {public_user.cluster}".replace(' ', '') in queries)
        self.assertTrue(
            f"""
            CREATE TABLE IF NOT EXISTS {public_user.database}.distributed_usage_metrics_processed_log ON CLUSTER {public_user.cluster} (
                `date` Date,
                `database` String,
                `host` String,
                `read_bytes` UInt64,
                `written_bytes` UInt64
            )
            ENGINE = Distributed('metrics', 'default', 'usage_metrics_log')
            """.replace('\n', '').replace(' ', '') in queries, queries)

        self.assertTrue(
            f"""
            CREATE MATERIALIZED VIEW IF NOT EXISTS {public_user.database}.usage_metrics_processed_log_view ON CLUSTER {public_user.cluster}
            TO {public_user.database}.distributed_usage_metrics_processed_log AS (
                SELECT
                    event_date as date,
                    current_database as database,
                    hostName() as host,
                    sum(read_bytes) as read_bytes,
                    sum(written_bytes) as written_bytes
                FROM system.query_log
                WHERE
                    type > 1
                    AND current_database not in ('default', 'system')
                    AND startsWith(http_user_agent, 'tb')
                    AND http_user_agent NOT IN ('tb-internal-query', 'tb-ui-query')
                GROUP BY host, database, date
            )
            """.replace('\n', '').replace(' ', '') in queries, queries)


class TestInitMetricsTables1(BaseTestInitTables):
    @patch.object(HTTPClient, 'query')
    # We disable the cluster checks in this test since we expect things to be created without cluster on purpose
    @patch.object(ClusterPatches, 'ENABLED', False)
    @tornado.testing.gen_test
    async def test_query_templates_for_init_metrics_tables_for_a_host(self, query_mock):
        empty_data_future = asyncio.Future()
        empty_data_future.set_result((None, '{"data": []}'))
        cluster_future = asyncio.Future()
        cluster_future.set_result((None,
                                   '{"data": [{"cluster": "cluster_1", "host_address": "127.0.0.1"},{"cluster": "cluster_2", "host_address": "127.0.0.2"}]}'))

        query_mock.side_effect = [
            empty_data_future,
            cluster_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future,
            empty_data_future
        ]

        public_user = public.get_public_user()
        public_user.clusters = ['internal']
        public_user.save()

        with patch.object(Users, 'add_datasource') as mock_add_datasource:
            await init_metrics_tables(
                host=public.get_public_user().database_server,
                metrics_cluster='metrics',
                metrics_database_server='http://127.0.0.1',
                metrics_cluster_tables=[],
                metrics_cluster_views=[],
                metrics_tables=DEFAULT_METRICS_TABLES,
                metrics_views=DEFAULT_METRICS_VIEWS,
                metrics_database='default',
                add_datasources=False
            )

        self.assertEqual(mock_add_datasource.call_count, 0)

        queries = [' '.join(call.args[0].split()).replace(' ', '') for call in query_mock.call_args_list]

        self.assertTrue(f"CREATE DATABASE IF NOT EXISTS {public_user.database}".replace(' ', '') in queries)
        self.assertTrue(
            f"""
            CREATE TABLE IF NOT EXISTS {public_user.database}.distributed_usage_metrics_processed_log (
                `date` Date,
                `database` String,
                `host` String,
                `read_bytes` UInt64,
                `written_bytes` UInt64
            )
            ENGINE = Distributed('metrics', 'default', 'usage_metrics_log')
            """.replace('\n', '').replace(' ', '') in queries, queries)

        self.assertTrue(
            f"""
                CREATE MATERIALIZED VIEW IF NOT EXISTS {public_user.database}.usage_metrics_processed_log_view
                TO {public_user.database}.distributed_usage_metrics_processed_log AS (
                    SELECT
                        event_date as date,
                        current_database as database,
                        hostName() as host,
                        sum(read_bytes) as read_bytes,
                        sum(written_bytes) as written_bytes
                    FROM system.query_log
                    WHERE
                        type > 1
                        AND current_database not in ('default', 'system')
                        AND startsWith(http_user_agent, 'tb')
                        AND http_user_agent NOT IN ('tb-internal-query', 'tb-ui-query')
                    GROUP BY host, database, date
                )
                """.replace('\n', '').replace(' ', '') in queries, queries)


class TestInitMetricsTables2(BaseTestInitTables):
    @tornado.testing.gen_test
    async def test_query_templates_metrics_tables_as_internal(self):
        await init_internal_tables(
            tables=DEFAULT_METRICS_TABLES,
            views=None,
            metrics_cluster='tinybird',
            metrics_database='default',
            job_executor=self.job_executor
        )
        self.created_tables += [table.name for table in DEFAULT_METRICS_TABLES]

        workspace = User.get_by_id(self.workspace.id)
        datasource = workspace.get_datasource('distributed_usage_metrics_processed_log')
        engine, _ = await datasource.table_metadata(self.workspace)
        self.assertEqual(engine.engine_full, "Distributed('tinybird', 'default', 'usage_metrics_log')")


class TestInitMetricsTables3(BaseTestInitTables):
    @tornado.testing.gen_test
    async def test_query_templates_metrics_cluster_tables_as_internal(self):
        await init_internal_tables(
            tables=DEFAULT_METRICS_CLUSTER_TABLES,
            views=None,
            metrics_cluster='tinybird',
            metrics_database='default',
            job_executor=self.job_executor
        )
        self.created_tables += [table.name for table in DEFAULT_METRICS_CLUSTER_TABLES]

        workspace = User.get_by_id(self.workspace.id)
        datasource = workspace.get_datasource('usage_metrics_log')
        engine, _ = await datasource.table_metadata(self.workspace)
        self.assertEqual(engine.engine_full, "SummingMergeTree() PARTITION BY toYYYYMM(date) ORDER BY (host, database, date)")


@pytest.mark.parametrize("uri,expected", [
    pytest.param("/v0/datasources/mine?query=value", True, id="If uri starts with /vX..., it comes from API"),
    pytest.param("/v32432/pipes", True, id="If uri starts with any version, it comes from API"),
    pytest.param("/internal/health", True, id="If uri starts with /internal, it is internal"),
    pytest.param("/login", False, id="Any other urls are not interal/api"),
    pytest.param(None, False, id="Empty url is not interal/api"),
])
def test_uri_is_internal_or_from_api(uri: Optional[str], expected: bool):
    result = uri_is_interna_or_from_api(uri)
    assert result == expected


class TestInternal(unittest.TestCase):
    def test_internal_workspace_plan(self):
        pu = public.get_public_user()
        self.assertEqual(pu.plan, BillingPlans.CUSTOM)




# This is a command file for our CLI. Please keep it clean.
#
# - If it makes sense and only when strictly necessary, you can create utility functions in this file.
# - But please, **do not** interleave utility functions and command definitions.

from typing import Any, Dict, List, Tuple
import click
from click import Context
import yaml

from tinybird.client import TinyB
from tinybird.tb_cli_modules.cli import cli
from tinybird.tb_cli_modules.common import coro, get_config_and_hosts, \
    create_workspace_branch, switch_workspace, switch_to_workspace_by_user_workspace_data, \
    print_current_workspace, _get_config, print_data_branch_summary, echo_safe_humanfriendly_tables_format_smart_table, \
    get_current_main_workspace, get_current_workspace_branches, MAIN_BRANCH, print_current_branch, \
    merge_workspace_branch, print_branch_regression_tests_summary
from tinybird.feedback_manager import FeedbackManager
from tinybird.datafile import wait_job
from tinybird.config import VERSION


@cli.group(hidden=True)
@click.pass_context
def branch(ctx: Context) -> None:
    """Branch commands.
    """


@branch.command(name="ls", hidden=True)
@click.pass_context
@coro
async def branch_ls(ctx: Context) -> None:
    """List all the branches from the workspace token.
    """

    obj: Dict[str, Any] = ctx.ensure_object(dict)
    client = obj['client']
    config = obj['config']

    if 'id' not in config:
        config = await _get_config(config['host'], config['token'], load_tb_file=False)

    current_main_workspace = await get_current_main_workspace(client, config)

    if current_main_workspace['id'] != config['id']:
        client = TinyB(current_main_workspace['token'], config['host'], version=VERSION)

    response = await client.branches()

    columns = ['name', 'id', 'current']
    table: List[Tuple[str, str, bool]] = [(MAIN_BRANCH, current_main_workspace['id'], config['id'] == current_main_workspace['id'])]

    for branch in response['branches']:
        table.append((branch['name'], branch['id'], config['id'] == branch['id']))

    await print_current_workspace(ctx)

    click.echo(FeedbackManager.info_branches())
    echo_safe_humanfriendly_tables_format_smart_table(table, column_names=columns)


@branch.command(name='use', hidden=True)
@click.argument('branch_name_or_id')
@click.pass_context
@coro
async def branch_use(
    ctx: Context,
    branch_name_or_id: str
) -> None:
    """Switch to another branch. Use 'tb branch ls' to list the branches you have access to.
    """

    client: TinyB = ctx.ensure_object(dict)['client']
    config, host, ui_host = await get_config_and_hosts(ctx)

    current_main_workspace = await get_current_main_workspace(client, config)
    if branch_name_or_id == MAIN_BRANCH:
        await switch_to_workspace_by_user_workspace_data(ctx, current_main_workspace)
    else:
        await switch_workspace(ctx, branch_name_or_id, only_branches=True)


@branch.command(name='current', hidden=True)
@click.pass_context
@coro
async def branch_current(ctx: Context) -> None:
    """Show the branch you're currently authenticated into.
    """

    await print_current_branch(ctx)


@branch.command(name='create', short_help="Create a new Branch from the Workspace you are authenticated", hidden=True)
@click.argument('branch_name', required=False)
@click.option('--last-partition', is_flag=True, default=False, help="When enabled, last modified partition is attached from the origin Workspace to the Branch")
@click.option('--all', is_flag=True, default=False, help="When enabled, all data from the origin Workspace is attached to the Branch. Use only if you actually need all the data in the branch.")
@click.option('--wait', is_flag=True, default=False, help="Waits for data branch jobs to finish, showing a progress bar. Disabled by default.")
@click.pass_context
@coro
async def create_branch(
    ctx: Context,
    branch_name: str,
    last_partition: bool,
    all: bool,
    wait: bool
) -> None:

    if last_partition and all:
        click.echo(FeedbackManager.error_exception(error="Use --last-partition or --all but not both"))
        return

    await create_workspace_branch(ctx, branch_name, last_partition, all, wait)


@branch.command(name='rm', short_help="Removes a Branch for your Tinybird user and it can't be recovered.", hidden=True)
@click.argument('branch_name_or_id')
@click.option('--yes', is_flag=True, default=False, help="Do not ask for confirmation")
@click.pass_context
@coro
async def delete_branch(
    ctx: Context,
    branch_name_or_id: str,
    yes: bool
) -> None:
    """Remove a branch where you are an admin.
    """

    client: TinyB = ctx.ensure_object(dict)['client']
    config, _, _ = await get_config_and_hosts(ctx)

    if branch_name_or_id == MAIN_BRANCH:
        click.echo(FeedbackManager.error_not_allowed_in_main_branch())
        return

    try:
        workspace_branches = await get_current_workspace_branches(client, config)
        workspace_to_delete = next((workspace for workspace in workspace_branches
                                    if workspace['name'] == branch_name_or_id or workspace['id'] == branch_name_or_id),
                                   None)
    except Exception as e:
        click.echo(FeedbackManager.error_exception(error=str(e)))
        return

    if not workspace_to_delete:
        raise click.ClickException(FeedbackManager.error_branch(branch=branch_name_or_id))

    if yes or click.confirm(FeedbackManager.warning_confirm_delete_branch(branch=workspace_to_delete['name'])):
        need_to_switch_to_main = workspace_to_delete.get('main') and config['id'] == workspace_to_delete['id']
        # get origin workspace if deleting current branch
        if need_to_switch_to_main:
            try:
                workspaces = (await client.user_workspaces()).get('workspaces', [])
                workspace_main = next((workspace for workspace in workspaces if
                                       workspace['id'] == workspace_to_delete['main']), None)
            except Exception:
                workspace_main = None
        try:
            await client.delete_branch(workspace_to_delete['id'])
            click.echo(FeedbackManager.success_branch_deleted(branch_name=workspace_to_delete['name']))
        except Exception as e:
            click.echo(FeedbackManager.error_exception(error=str(e)))
            return
        else:
            if need_to_switch_to_main:
                if workspace_main:
                    await switch_to_workspace_by_user_workspace_data(ctx, workspace_main)
                else:
                    click.echo(FeedbackManager.error_switching_to_main())


@branch.command(name='data', short_help="Perform a data branch operation, see flags for details", hidden=True)
@click.option('--last-partition', is_flag=True, default=False, help="When enabled, last modified partition is attached from the origin Workspace to the Branch")
@click.option('--all', is_flag=True, default=False, help="When enabled, all data from the origin Workspace is attached to the Branch. Use only if you actually need all the data in the branch.")
@click.option('--wait', is_flag=True, default=False, help="Waits for data branch jobs to finish, showing a progress bar. Disabled by default.")
@click.pass_context
@coro
async def data_branch(
    ctx: Context,
    last_partition: bool,
    all: bool,
    wait: bool
) -> None:

    if last_partition and all:
        click.echo(FeedbackManager.error_exception(error="Use --last-partition or --all but not both"))
        return

    if not last_partition and not all:
        click.echo(FeedbackManager.error_exception(error="Use --last-partition or --all"))
        return

    obj: Dict[str, Any] = ctx.ensure_object(dict)
    client = obj['client']
    config = obj['config']

    current_main_workspace = await get_current_main_workspace(client, config)
    if current_main_workspace['id'] == config['id']:
        click.echo(FeedbackManager.error_not_allowed_in_main_branch())
        return

    try:
        response = await client.branch_workspace_data(config['id'], last_partition, all)
        if all:
            if 'job' not in response:
                raise click.ClickException(response)
            job_id = response['job']['job_id']
            job_url = response['job']['job_url']
            click.echo(FeedbackManager.info_data_branch_job_url(url=job_url))
            if wait:
                await wait_job(client, job_id, job_url, 'Data Branching')
                await print_data_branch_summary(client, job_id)
        else:
            await print_data_branch_summary(client, None, response)
            click.echo(FeedbackManager.success_workspace_data_branch())
    except Exception as e:
        click.echo(FeedbackManager.error_exception(error=str(e)))
        return


@branch.group("regression-tests", invoke_without_command=True)
@click.option('-f', '--filename', type=click.Path(exists=True), required=False, help="yaml file with regression-tests definition")
@click.option('--wait', is_flag=True, default=False, help="Waits for regression job to finish, showing a progress bar. Disabled by default.")
@click.pass_context
@coro
async def regression_tests(ctx, filename: str, wait: bool):
    '''Branch regression-tests commands'''
    if filename:
        with open(filename, 'r') as file:
            regression_tests_commands = yaml.safe_load(file)
        client: TinyB = ctx.ensure_object(dict)['client']
        config = ctx.ensure_object(dict)['config']

        current_main_workspace = await get_current_main_workspace(client, config)
        if current_main_workspace['id'] == config['id']:
            click.echo(FeedbackManager.error_not_allowed_in_main_branch())
            return
        try:
            response = await client.branch_regression_tests_file(config['id'], regression_tests_commands)
            if 'job' not in response:
                raise click.ClickException(response)
            job_id = response['job']['job_id']
            job_url = response['job']['job_url']
            click.echo(FeedbackManager.info_regression_tests_branch_job_url(url=job_url))
            if wait:
                await wait_job(client, job_id, job_url, 'Regression tests')
                await print_branch_regression_tests_summary(client, job_id, config['host'])
        except click.ClickException as e:
            raise e
        except Exception as e:
            click.echo(FeedbackManager.error_exception(error=str(e)))
            return


@regression_tests.command(name='coverage', short_help="Run regression tests using coverage requests for branch vs main. It creates a regression-tests job", hidden=True)
@click.argument('pipe_name', required=False)
@click.option('--sample-by-params', type=click.IntRange(1, 100), default=1, required=False, help="When set, we will aggregate the pipe_stats_rt requests by extractURLParameterNames(assumeNotNull(url)) and for each combination we will take a sample of N requests")
@click.option('-m', '--match', multiple=True, required=False, help="Filter the checker requests by specific parameter. You can pass multiple parameters -m foo -m bar")
@click.option('--only-response-times', is_flag=True, default=False, help="Checks only response times")
@click.option('-ff', '--failfast', is_flag=True, default=False, help="When set, the checker will exit as soon one test fails")
@click.option('--ignore-order', is_flag=True, default=False, help="When set, the checker will ignore the order of list properties")
@click.option('--validate-processed-bytes', is_flag=True, default=False, help="When set, the checker will validate that the new version doesn't process more than 25% than the current version")
@click.option('--wait', is_flag=True, default=False, help="Waits for regression job to finish, showing a progress bar. Disabled by default.")
@click.pass_context
@coro
async def coverage(ctx: Context, pipe_name: str, sample_by_params: int, match: List[str],
                   only_response_times: bool, failfast: bool, ignore_order: bool, validate_processed_bytes: bool, wait: bool):
    client: TinyB = ctx.ensure_object(dict)['client']
    config = ctx.ensure_object(dict)['config']

    current_main_workspace = await get_current_main_workspace(client, config)
    if current_main_workspace['id'] == config['id']:
        click.echo(FeedbackManager.error_not_allowed_in_main_branch())
        return
    try:
        response = await client.branch_regression_tests(config['id'], [pipe_name],
                                                        'coverage',
                                                        samples_by_params=sample_by_params,
                                                        matches=match if match else [],
                                                        only_response_times=only_response_times,
                                                        failfast=failfast,
                                                        ignore_order=ignore_order, validate_processed_bytes=validate_processed_bytes)
        if 'job' not in response:
            raise click.ClickException(response)
        job_id = response['job']['job_id']
        job_url = response['job']['job_url']
        click.echo(FeedbackManager.info_regression_tests_branch_job_url(url=job_url))
        if wait:
            await wait_job(client, job_id, job_url, 'Regression tests')
            await print_branch_regression_tests_summary(client, job_id, config['host'])
    except click.ClickException as e:
        raise e
    except Exception as e:
        click.echo(FeedbackManager.error_exception(error=str(e)))
        return


@regression_tests.command(name='last', short_help="Run regression tests using last requests for branch vs main. It creates a regression-tests job", hidden=True)
@click.argument('pipe_name', required=False)
@click.option('-l', '--limit', type=click.IntRange(0, 100), default=0, required=False, help="Number of requests to validate")
@click.option('-m', '--match', multiple=True, required=False, help="Filter the checker requests by specific parameter. You can pass multiple parameters -m foo -m bar")
@click.option('--only-response-times', is_flag=True, default=False, help="Checks only response times")
@click.option('-ff', '--failfast', is_flag=True, default=False, help="When set, the checker will exit as soon one test fails")
@click.option('--ignore-order', is_flag=True, default=False, help="When set, the checker will ignore the order of list properties")
@click.option('--validate-processed-bytes', is_flag=True, default=False, help="When set, the checker will validate that the new version doesn't process more than 25% than the current version")
@click.option('--wait', is_flag=True, default=False, help="Waits for regression job to finish, showing a progress bar. Disabled by default.")
@click.pass_context
@coro
async def last(ctx: Context, pipe_name: str, limit: int, match: List[str],
               only_response_times: bool, failfast: bool, ignore_order: bool, validate_processed_bytes: bool, wait: bool):
    client: TinyB = ctx.ensure_object(dict)['client']
    config = ctx.ensure_object(dict)['config']

    current_main_workspace = await get_current_main_workspace(client, config)
    if current_main_workspace['id'] == config['id']:
        click.echo(FeedbackManager.error_not_allowed_in_main_branch())
        return
    try:
        response = await client.branch_regression_tests(config['id'], [pipe_name],
                                                        'last',
                                                        limit=limit,
                                                        matches=match if match else [],
                                                        only_response_times=only_response_times,
                                                        failfast=failfast,
                                                        ignore_order=ignore_order, validate_processed_bytes=validate_processed_bytes)
        if 'job' not in response:
            raise click.ClickException(response)
        job_id = response['job']['job_id']
        job_url = response['job']['job_url']
        click.echo(FeedbackManager.info_regression_tests_branch_job_url(url=job_url))
        if wait:
            await wait_job(client, job_id, job_url, 'Regression tests')
            await print_branch_regression_tests_summary(client, job_id, config['host'])
    except click.ClickException as e:
        raise e
    except Exception as e:
        click.echo(FeedbackManager.error_exception(error=str(e)))
        return


@regression_tests.command(name='manual', short_help="Run regression tests using manual requests for branch vs main. It creates a regression-tests job", hidden=True,
                          context_settings=dict(allow_extra_args=True, ignore_unknown_options=True))
@click.argument('pipe_name', required=False)
@click.option('--only-response-times', is_flag=True, default=False, help="Checks only response times")
@click.option('-ff', '--failfast', is_flag=True, default=False, help="When set, the checker will exit as soon one test fails")
@click.option('--ignore-order', is_flag=True, default=False, help="When set, the checker will ignore the order of list properties")
@click.option('--validate-processed-bytes', is_flag=True, default=False, help="When set, the checker will validate that the new version doesn't process more than 25% than the current version")
@click.option('--wait', is_flag=True, default=False, help="Waits for regression job to finish, showing a progress bar. Disabled by default.")
@click.pass_context
@coro
async def manual(ctx: Context, pipe_name: str, only_response_times: bool,
                 failfast: bool, ignore_order: bool, validate_processed_bytes: bool, wait: bool):
    client: TinyB = ctx.ensure_object(dict)['client']
    config = ctx.ensure_object(dict)['config']

    params = [{ctx.args[i][2:]: ctx.args[i + 1] for i in range(0, len(ctx.args), 2)}]

    current_main_workspace = await get_current_main_workspace(client, config)
    if current_main_workspace['id'] == config['id']:
        click.echo(FeedbackManager.error_not_allowed_in_main_branch())
        return
    try:
        response = await client.branch_regression_tests(config['id'], [pipe_name],
                                                        'manual',
                                                        params=params if params else [],
                                                        only_response_times=only_response_times,
                                                        failfast=failfast,
                                                        ignore_order=ignore_order, validate_processed_bytes=validate_processed_bytes)
        if 'job' not in response:
            raise click.ClickException(response)
        job_id = response['job']['job_id']
        job_url = response['job']['job_url']
        click.echo(FeedbackManager.info_regression_tests_branch_job_url(url=job_url))
        if wait:
            await wait_job(client, job_id, job_url, 'Regression tests')
            await print_branch_regression_tests_summary(client, job_id, config['host'])
    except click.ClickException as e:
        raise e
    except Exception as e:
        click.echo(FeedbackManager.error_exception(error=str(e)))
        return


@branch.command(name='merge', short_help="Merge a Branch to main. It creates a deployment job", hidden=True)
@click.option('--wait', is_flag=True, default=False, help="Waits for data branch jobs to finish, showing a progress bar. Disabled by default.")
@click.option('--verbose', is_flag=True, default=False, help="Print DEBUG logs.")
@click.pass_context
@coro
async def merge_branch(
    ctx: Context,
    wait: bool,
    verbose: bool
) -> None:

    client: TinyB = ctx.ensure_object(dict)['client']
    config = ctx.ensure_object(dict)['config']

    current_main_workspace = await get_current_main_workspace(client, config)
    if current_main_workspace['id'] == config['id']:
        click.echo(FeedbackManager.error_not_allowed_in_main_branch())
        return

    await merge_workspace_branch(ctx, config['id'], current_main_workspace, wait, verbose)


@branch.group()
@click.pass_context
def datasource(ctx: Context) -> None:
    """Branch data source commands.
    """


@datasource.command(name="copy")
@click.argument('datasource_name')
@click.option('--sql', default=None, help='SQL query to copy', hidden=True, required=False)
@click.option('--sql-from-main', is_flag=True, default=False, help='SQL query from main to copy', hidden=True, required=False)
@click.option('--wait', is_flag=True, default=False, help="Wait for copy job to finish, disabled by default")
@click.pass_context
@coro
async def datasource_copy_from_main(
    ctx: Context,
    datasource_name: str,
    sql: str,
    sql_from_main: bool,
    wait: bool
) -> None:
    """Copy data source from main.
    """

    client: TinyB = ctx.ensure_object(dict)['client']
    config = ctx.ensure_object(dict)['config']

    if sql and sql_from_main:
        click.echo(FeedbackManager.error_exception(error="Use --sql or --sql-from-main but not both"))
        return

    if not sql and not sql_from_main:
        click.echo(FeedbackManager.error_exception(error="Use --sql or --sql-from-main"))
        return

    current_main_workspace = await get_current_main_workspace(client, config)
    if current_main_workspace['id'] == config['id']:
        click.echo(FeedbackManager.error_not_allowed_in_main_branch())
        return

    response = await client.datasource_query_copy(datasource_name, sql if sql else f"SELECT * FROM main.{datasource_name}")
    if 'job' not in response:
        raise click.ClickException(response)
    job_id = response['job']['job_id']
    job_url = response['job']['job_url']
    if sql:
        click.echo(FeedbackManager.info_copy_with_sql_job_url(sql=sql,
                                                              datasource_name=datasource_name,
                                                              url=job_url))
    else:
        click.echo(FeedbackManager.info_copy_from_main_job_url(datasource_name=datasource_name, url=job_url))
    if wait:
        base_msg = 'Copy from main' if sql_from_main else f'Copy from {sql}'
        await wait_job(client, job_id, job_url, f"{base_msg} to {datasource_name}")



import uuid
import pytest
from datetime import datetime, timedelta, date
from urllib.parse import urlencode
from tests.views.base_test import STRIPE_SUBSCRIPTION_MOCK, BaseTest, TBApiProxy, UsageMetricsTestMixin
from unittest.mock import patch, call, AsyncMock

from tinybird.plans import PlansService, DEFAULT_PLAN_CONFIG, PlanConfigConcepts, BuildPlanTracker
from tinybird.token_scope import scopes
from tinybird.user import User, UserAccount, public
from tinybird.constants import BillingPlans
from tinybird.views.mailgun import MailgunService, NotificationResponse
from tests.utils import exec_sql

from .conftest import is_main_process


STORAGE_BYTES = int(10.5 * (1000 ** 3))
READ_BYTES_TODAY = int(10 * (1000 ** 3))
WRITTEN_BYTES_TODAY = int(10.5 * (1000 ** 3))
READ_BYTES_YESTERDAY = int(20 * (1000 ** 3))
WRITTEN_BYTES_YESTERDAY = int(20.5 * (1000 ** 3))


STRIPE_SUBSCRIPTIONS_MOCK = {'data': [STRIPE_SUBSCRIPTION_MOCK], "has_more": False}


class TestTrackUsageRecords(BaseTest, UsageMetricsTestMixin):

    def setUp(self):
        super(TestTrackUsageRecords, self).setUp()
        self.subscriptions_to_delete = []
        self.token = UserAccount.get_by_id(self.USER_ID).get_token_for_scope(scopes.AUTH)
        self.processed_item_id = STRIPE_SUBSCRIPTION_MOCK['items']['data'][0]['id']
        self.storage_item_id = STRIPE_SUBSCRIPTION_MOCK['items']['data'][1]['id']
        self.tb_api = TBApiProxy(self)
        self.now = datetime.utcnow()
        self.today = self.now.date()
        self.yesterday = self.today - timedelta(days=1)

    def assert_usage_records(self, processed_quantity_today, storage_quantity, stripe_create_usage_record_mock,
                             timestamp_today=None,
                             timestamp_yesterday=None,
                             processed_quantity_yesterday=None):
        assert stripe_create_usage_record_mock.call_count == 2 if processed_quantity_yesterday is None else 3
        if timestamp_today is None:
            timestamp_today = datetime.combine(self.today, datetime.min.time())

        if timestamp_yesterday is None:
            timestamp_yesterday = datetime.combine(self.today, datetime.min.time()) - timedelta(days=1)

        create_usage_calls_calls = [call(id=self.processed_item_id,
                                         action='set', timestamp=timestamp_today,
                                         quantity=processed_quantity_today),
                                    call(id=self.storage_item_id,
                                         action='set', timestamp=timestamp_today,
                                         quantity=storage_quantity)]
        if processed_quantity_yesterday:
            create_usage_calls_calls.append(call(id=self.processed_item_id,
                                                 action='set', timestamp=timestamp_yesterday,
                                                 quantity=processed_quantity_yesterday))

        stripe_create_usage_record_mock.assert_has_calls(create_usage_calls_calls, any_order=True)

    @patch('stripe.SubscriptionItem.create_usage_record', return_value={})
    def test_track_usage_records_subscription_from_yesterday(
            self,
            stripe_create_usage_record_mock,
    ):
        ws_name = f'workspace_track_{uuid.uuid4().hex}'
        workspace = self.tb_api.create_workspace(self.token, ws_name)
        workspace = User.get_by_id(workspace['id'])
        self.workspaces_to_delete.append(workspace)

        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace,
                                                                        datetime.combine(self.today,
                                                                                         datetime.min.time()),
                                                                        10000, STORAGE_BYTES, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.insert_row_processed_metrics(workspace, self.today, READ_BYTES_TODAY,
                                          WRITTEN_BYTES_TODAY)
        self.insert_row_processed_metrics(workspace, self.yesterday, READ_BYTES_YESTERDAY,
                                          WRITTEN_BYTES_YESTERDAY)

        self.tb_api.create_customer(self.token, workspace['id'])
        self.tb_api.setup_payment_intent(self.token, workspace['id'])
        self.tb_api.subscribe_to_pro(self.token, workspace['id'])
        workspace = User.get_by_id(workspace['id'])

        subscription = STRIPE_SUBSCRIPTION_MOCK
        subscription_datetime = self.now - timedelta(days=1)
        subscription["created"] = datetime.timestamp(subscription_datetime)
        subscription["current_period_start"] = datetime.timestamp(subscription_datetime)

        metrics_cluster = "metrics"

        PlansService.track_usage_records(
            workspace,
            subscription,
            metrics_cluster)

        self.assert_usage_records(20, 10, stripe_create_usage_record_mock,
                                  processed_quantity_yesterday=40,
                                  timestamp_yesterday=subscription_datetime + timedelta(seconds=1))

    @patch('stripe.SubscriptionItem.create_usage_record', return_value={})
    def test_track_usage_records_same_day_of_subscription(
            self,
            stripe_create_usage_record_mock,
    ):
        ws_name = f'workspace_track_{uuid.uuid4().hex}'
        workspace = self.tb_api.create_workspace(self.token, ws_name)
        workspace = User.get_by_id(workspace['id'])
        self.workspaces_to_delete.append(workspace)

        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace,
                                                                        datetime.combine(self.today,
                                                                                         datetime.min.time()),
                                                                        10000, STORAGE_BYTES, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.insert_row_processed_metrics(workspace, self.today, READ_BYTES_TODAY,
                                          WRITTEN_BYTES_TODAY)
        self.insert_row_processed_metrics(workspace, self.yesterday, READ_BYTES_YESTERDAY,
                                          WRITTEN_BYTES_YESTERDAY)

        self.tb_api.create_customer(self.token, workspace['id'])
        self.tb_api.setup_payment_intent(self.token, workspace['id'])
        self.tb_api.subscribe_to_pro(self.token, workspace['id'])
        workspace = User.get_by_id(workspace['id'])

        subscription = STRIPE_SUBSCRIPTION_MOCK
        subscription_timestamp = self.now
        subscription["created"] = datetime.timestamp(subscription_timestamp)
        subscription["current_period_start"] = datetime.timestamp(subscription_timestamp)

        metrics_cluster = "metrics"

        PlansService.track_usage_records(
            workspace,
            subscription,
            metrics_cluster)

        self.assert_usage_records(20, 10, stripe_create_usage_record_mock,
                                  timestamp_today=subscription_timestamp + timedelta(seconds=1))

    @patch('stripe.SubscriptionItem.create_usage_record', return_value={})
    def test_track_usage_records_subscription_from_days_before(
            self,
            stripe_create_usage_record_mock,
    ):
        ws_name = f'workspace_track_{uuid.uuid4().hex}'
        workspace = self.tb_api.create_workspace(self.token, ws_name)
        workspace = User.get_by_id(workspace['id'])
        self.workspaces_to_delete.append(workspace)

        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace,
                                                                        datetime.combine(self.today,
                                                                                         datetime.min.time()),
                                                                        10000, STORAGE_BYTES, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.insert_row_processed_metrics(workspace, self.today, READ_BYTES_TODAY,
                                          WRITTEN_BYTES_TODAY)
        self.insert_row_processed_metrics(workspace, self.yesterday, READ_BYTES_YESTERDAY,
                                          WRITTEN_BYTES_YESTERDAY)

        self.tb_api.create_customer(self.token, workspace['id'])
        self.tb_api.setup_payment_intent(self.token, workspace['id'])
        self.tb_api.subscribe_to_pro(self.token, workspace['id'])
        workspace = User.get_by_id(workspace['id'])

        subscription = STRIPE_SUBSCRIPTION_MOCK
        subscription_datetime = self.now - timedelta(days=2)
        subscription["created"] = datetime.timestamp(subscription_datetime)
        subscription["current_period_start"] = datetime.timestamp(subscription_datetime)

        metrics_cluster = "metrics"

        PlansService.track_usage_records(
            workspace,
            subscription,
            metrics_cluster)

        self.assert_usage_records(20, 10, stripe_create_usage_record_mock,
                                  processed_quantity_yesterday=40,
                                  timestamp_yesterday=datetime.combine(self.yesterday, datetime.min.time()))

    @patch('stripe.SubscriptionItem.create_usage_record', return_value={})
    def test_track_usage_records_billing_period_starts_today(
            self,
            stripe_create_usage_record_mock,
    ):
        ws_name = f'workspace_track_{uuid.uuid4().hex}'
        workspace = self.tb_api.create_workspace(self.token, ws_name)
        workspace = User.get_by_id(workspace['id'])
        self.workspaces_to_delete.append(workspace)

        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace,
                                                                        datetime.combine(self.today,
                                                                                         datetime.min.time()),
                                                                        10000, STORAGE_BYTES, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.insert_row_processed_metrics(workspace, self.today, READ_BYTES_TODAY,
                                          WRITTEN_BYTES_TODAY)
        self.insert_row_processed_metrics(workspace, self.yesterday, READ_BYTES_YESTERDAY,
                                          WRITTEN_BYTES_YESTERDAY)

        self.tb_api.create_customer(self.token, workspace['id'])
        self.tb_api.setup_payment_intent(self.token, workspace['id'])
        self.tb_api.subscribe_to_pro(self.token, workspace['id'])
        workspace = User.get_by_id(workspace['id'])

        subscription = STRIPE_SUBSCRIPTION_MOCK
        subscription_datetime = self.now - timedelta(days=2)
        subscription["created"] = datetime.timestamp(subscription_datetime)
        subscription["current_period_start"] = datetime.timestamp(datetime.combine(self.today, datetime.min.time()))

        metrics_cluster = "metrics"

        PlansService.track_usage_records(
            workspace,
            subscription,
            metrics_cluster)

        self.assert_usage_records(20, 10, stripe_create_usage_record_mock,
                                  timestamp_today=datetime.fromtimestamp(subscription["current_period_start"] + 1))

    @patch('stripe.SubscriptionItem.create_usage_record', return_value={})
    def test_track_usage_records_billing_period_starts_yesterday(
            self,
            stripe_create_usage_record_mock,
    ):
        ws_name = f'workspace_track_{uuid.uuid4().hex}'
        workspace = self.tb_api.create_workspace(self.token, ws_name)
        workspace = User.get_by_id(workspace['id'])
        self.workspaces_to_delete.append(workspace)

        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace,
                                                                        datetime.combine(date.today(),
                                                                                         datetime.min.time()),
                                                                        10000, STORAGE_BYTES, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.insert_row_processed_metrics(workspace, self.today, READ_BYTES_TODAY,
                                          WRITTEN_BYTES_TODAY)
        self.insert_row_processed_metrics(workspace, self.yesterday, READ_BYTES_YESTERDAY,
                                          WRITTEN_BYTES_YESTERDAY)

        self.tb_api.create_customer(self.token, workspace['id'])
        self.tb_api.setup_payment_intent(self.token, workspace['id'])
        self.tb_api.subscribe_to_pro(self.token, workspace['id'])
        workspace = User.get_by_id(workspace['id'])

        subscription = STRIPE_SUBSCRIPTION_MOCK
        subscription_datetime = self.now - timedelta(days=3)
        subscription["created"] = datetime.timestamp(subscription_datetime)
        subscription["current_period_start"] = datetime.timestamp(datetime.combine(self.today,
                                                                                   datetime.min.time()) - timedelta(days=1))

        metrics_cluster = "metrics"

        PlansService.track_usage_records(
            workspace,
            subscription,
            metrics_cluster)

        self.assert_usage_records(20, 10, stripe_create_usage_record_mock,
                                  processed_quantity_yesterday=40,
                                  timestamp_yesterday=datetime.fromtimestamp(subscription["current_period_start"] + 1))


@pytest.mark.serial  # Modifies the public user
@pytest.mark.skipif(not is_main_process(), reason="Serial test")
class TestTrackBuildPlanLimits(BaseTest, UsageMetricsTestMixin):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.STORAGE_BYTES_EXCEEDED = int(11 * (1000 ** 3))
        cls.STORAGE_BYTES_REACHING = int(8 * (1000 ** 3))
        cls.STORAGE_BYTES_IN_PLAN = int(7 * (1000 ** 3))

    def tearDown(self):
        public_user = public.get_public_user()
        token = public_user.get_token_for_scope(scopes.ADMIN)
        self.tb_api.truncate_datasource(token, 'workspaces_all')
        super().tearDown()

    def setUp(self):
        super(TestTrackBuildPlanLimits, self).setUp()
        self.rand_id = f"{uuid.uuid4().hex}"
        self.user = UserAccount.register(f'test_track_build_plan_limits{self.rand_id}@example.com', 'pass')
        self.users_to_delete.append(self.user)
        self.token = self.user.get_token_for_scope(scopes.AUTH)
        self.tb_api = TBApiProxy(self)
        self.public = public.get_public_user()
        self.pipe_stats = self.public.get_datasource('pipe_stats')
        self.spans = self.public.get_datasource('spans')

    def exec_workspace_datatracker(self, workspaces):
        from tinybird.job import WorkspaceDatabaseUsageTracker
        data_tracker = WorkspaceDatabaseUsageTracker()
        data_tracker.track_model(workspaces, 'workspaces_all', ['id', 'name', 'database', 'database_server', 'plan',
                                                                'deleted', 'created_at'])
        self.wait_for_public_table_replication('workspaces_all')

    def _dummy_query_api(self, token, times=1):
        params = {
            'q': 'select 1',
            'token': token
        }
        url = f'/v0/sql?{urlencode(params)}'
        for _ in range(1, times + 1):
            response = self.fetch(url)
            self.assertEqual(response.code, 200, response.body)
        self.force_flush_of_span_records()
        self.wait_for_public_table_replication('pipe_stats')

    def assertPipeStats(self, workspace, count):
        pipe_stats_result = exec_sql(self.public.database, f"select sum(view_count) as c from {self.public.database}.{self.pipe_stats.id} where user_id='{workspace.id}' AND pipe_id='query_api' FORMAT JSON")
        self.assertEqual(int(pipe_stats_result['data'][0]['c']), count, exec_sql(self.public.database, f"select * from {self.public.database}.{self.pipe_stats.id} where user_id='{workspace.id}' AND pipe_id='query_api' FORMAT JSON"))

    def assertSpans(self, workspace, count):
        pipe_stats_result = exec_sql(self.public.database, f"select count() as c from {self.public.database}.{self.spans.id} where workspace='{workspace.id}' AND operation_name = 'APIQueryHandler' FORMAT JSON")
        self.assertEqual(int(pipe_stats_result['data'][0]['c']), count, exec_sql(self.public.database, f"select * from {self.public.database}.{self.spans.id} where workspace='{workspace.id}' AND operation_name = 'APIQueryHandler' FORMAT JSON"))

    def test_track_build_plan_limits(self):
        workspace_exceeded = self.tb_api.create_workspace(self.token, f'workspace_exceeded_{self.rand_id}')
        workspace_exceeded = User.get_by_id(workspace_exceeded['id'])
        exceeded_token = workspace_exceeded.get_token_for_scope(scopes.ADMIN)
        self._dummy_query_api(exceeded_token, 2)
        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace_exceeded,
                                                                        datetime.combine(date.today(),
                                                                                         datetime.min.time()),
                                                                        10000, self.STORAGE_BYTES_EXCEEDED, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.workspaces_to_delete.append(workspace_exceeded)
        self.assertSpans(workspace_exceeded, 2)
        self.assertPipeStats(workspace_exceeded, 2)

        workspace_reaching = self.tb_api.create_workspace(self.token, f'workspace_reaching_{self.rand_id}')
        workspace_reaching = User.get_by_id(workspace_reaching['id'])
        reaching_token = workspace_reaching.get_token_for_scope(scopes.ADMIN)
        self._dummy_query_api(reaching_token, 1)
        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace_reaching,
                                                                        datetime.combine(date.today(),
                                                                                         datetime.min.time()),
                                                                        10000, self.STORAGE_BYTES_REACHING, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.workspaces_to_delete.append(workspace_reaching)
        self.assertSpans(workspace_reaching, 1)
        self.assertPipeStats(workspace_reaching, 1)

        workspace_in_plan = self.tb_api.create_workspace(self.token, f'workspace_in_plan_{self.rand_id}')
        workspace_in_plan = User.get_by_id(workspace_in_plan['id'])
        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace_in_plan,
                                                                        datetime.combine(date.today(),
                                                                                         datetime.min.time()),
                                                                        10000, self.STORAGE_BYTES_IN_PLAN, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.workspaces_to_delete.append(workspace_in_plan)
        self.assertSpans(workspace_in_plan, 0)
        self.assertPipeStats(workspace_in_plan, 0)

        workspace_not_in_build_plan = self.tb_api.create_workspace(self.token, f'workspace_not_in_build_plan_{self.rand_id}')
        workspace_not_in_build_plan = User.get_by_id(workspace_not_in_build_plan['id'])
        workspace_not_in_build_plan.plan = BillingPlans.PRO
        workspace_not_in_build_plan.save()
        not_in_build_plan_token = workspace_not_in_build_plan.get_token_for_scope(scopes.ADMIN)
        self._dummy_query_api(not_in_build_plan_token, 2)
        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace_not_in_build_plan,
                                                                        datetime.combine(date.today(),
                                                                                         datetime.min.time()),
                                                                        10000, self.STORAGE_BYTES_EXCEEDED, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.workspaces_to_delete.append(workspace_not_in_build_plan)
        self.assertSpans(workspace_not_in_build_plan, 2)
        self.assertPipeStats(workspace_not_in_build_plan, 2)

        self.exec_workspace_datatracker([workspace_exceeded, workspace_reaching, workspace_in_plan])

        # mock requests limit
        BuildPlanTracker.MAX_API_REQUESTS_PER_DAY_LIMIT = 1

        MailgunService.send_notification_on_build_plan_limits = AsyncMock(return_value=NotificationResponse(200))

        BuildPlanTracker.track_limits()

        BuildPlanTracker.MAX_API_REQUESTS_PER_DAY_LIMIT = DEFAULT_PLAN_CONFIG[PlanConfigConcepts.DEV_MAX_API_REQUESTS_PER_DAY]

        send_notification_calls = [call([self.user.email], workspace_reaching.id, workspace_reaching.name, 1, 1, 8, 10, 0.07, 0.34, exceeded=False),
                                   call([self.user.email], workspace_exceeded.id, workspace_exceeded.name, 2, 1, 11, 10, 0.07, 0.34, exceeded=True)]

        self.assertEqual(MailgunService.send_notification_on_build_plan_limits.call_count, 2)
        MailgunService.send_notification_on_build_plan_limits.assert_has_calls(send_notification_calls, any_order=True)

    def test_track_build_plan_limits_just_requests(self):
        workspace_reaching = self.tb_api.create_workspace(self.token, f'workspace_reaching_{self.rand_id}')
        workspace_reaching = User.get_by_id(workspace_reaching['id'])
        reaching_token = workspace_reaching.get_token_for_scope(scopes.ADMIN)
        self._dummy_query_api(reaching_token, 1)
        self.insert_row_storage_metrics_using_usage_metrics_storage__v2(workspace_reaching,
                                                                        datetime.combine(date.today(),
                                                                                         datetime.min.time()),
                                                                        10000, self.STORAGE_BYTES_IN_PLAN, 0, 0)
        self.wait_for_public_table_replication('usage_metrics_storage__v2')
        self.workspaces_to_delete.append(workspace_reaching)
        self.assertSpans(workspace_reaching, 1)
        self.assertPipeStats(workspace_reaching, 1)

        self.exec_workspace_datatracker([workspace_reaching])

        # mock requests limit
        BuildPlanTracker.MAX_API_REQUESTS_PER_DAY_LIMIT = 1

        MailgunService.send_notification_on_build_plan_limits = AsyncMock(return_value=NotificationResponse(200))

        BuildPlanTracker.track_limits()

        BuildPlanTracker.MAX_API_REQUESTS_PER_DAY_LIMIT = DEFAULT_PLAN_CONFIG[
            PlanConfigConcepts.DEV_MAX_API_REQUESTS_PER_DAY]

        send_notification_calls = [
            call([self.user.email], workspace_reaching.id, workspace_reaching.name, 1, 1, None, 10, 0.07, 0.34,
                 exceeded=False)]

        self.assertEqual(MailgunService.send_notification_on_build_plan_limits.call_count, 1)
        MailgunService.send_notification_on_build_plan_limits.assert_has_calls(send_notification_calls, any_order=True)



from setuptools import setup, find_packages
import os

# https://stackoverflow.com/a/59969843
try:
    from pip._internal.req import parse_requirements
except ImportError:
    # pip < 10
    from pip.req import parse_requirements


def package_files(directory):
    paths = []
    for (path, _directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


def requirements_from_file(path):
    try:
        # Attemps to use the old API
        parsed_requirements = parse_requirements(path, session=False)
        requirements = [str(ir.req) for ir in parsed_requirements]
    except Exception:
        # Recreate the generator as we consumed one item from it
        parsed_requirements = parse_requirements(path, session=False)
        requirements = [str(ir.requirement) for ir in parsed_requirements]
    return requirements


extra_files = package_files('tinybird/templates') + package_files('tinybird/static') + \
    package_files('tinybird/default_tables') + package_files('tinybird/default_pipes') + \
    package_files('tinybird/sql') + package_files('tinybird/scripts')


setup(
    name='tinybird',
    version='1.0',
    description='tinybird',
    long_description='tinybird analytics',
    long_description_content_type='text/markdown',
    author='tinybird.co',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    install_requires=requirements_from_file('requirements.txt'),
    package_dir={"tinybird": "tinybird/"},
    package_data={
        # 'tinybird': ['../templates/*', '../static/*']
        'tinybird': extra_files
    },
    setup_requires=[
        'pytest-runner',
        'cffi==1.14.5'
    ],
    cffi_modules=[
        "tinybird/fast_leb128_build.py:ffibuilder",
        "tinybird/csv_importer_find_endline_build.py:ffibuilder"],  # "filename:global"
    extras_require={
        'test': requirements_from_file('requirements-test.txt') + requirements_from_file('requirements-linters.txt'),
        'deploy': requirements_from_file('requirements-deploy.txt'),
        'doc': requirements_from_file('requirements-doc.txt'),
        'devtools': requirements_from_file('requirements-devtools.txt'),
    },
    entry_points={
        'console_scripts': [
            'tinybird_server=tinybird.app:run',
            'tinybird_tool=tinybird.cli:cli',
            'tb=tinybird.tb_cli:cli',
            'colibri=tinybird.colibri.main:cli',
        ],
    }
)



import tinybird.user
from tinybird.table import alter_table_operations
from tinybird.user import Users
from .views.base_test import BaseTest

import tornado.testing


class TestAlterSchemaValidation(BaseTest):

    def setUp(self):
        super(TestAlterSchemaValidation, self).setUp()
        self.workspace = Users.get_by_id(self.WORKSPACE_ID)

    @tornado.testing.gen_test
    async def test_same_schema(self):
        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = schema_a
        operations, operations_quarantine = await alter_table_operations(self.workspace, schema_a, schema_b)
        self.assertEqual(operations, [])
        self.assertEqual(operations_quarantine, [])

    @tornado.testing.gen_test
    async def test_add_column(self):
        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32,
            d String
        """
        operations, operations_quarantine = await alter_table_operations(self.workspace, schema_a, schema_b)
        self.assertEqual(operations, ['ADD COLUMN `d` String'])
        self.assertEqual(operations_quarantine, ['ADD COLUMN `d` Nullable(String)'])

    @tornado.testing.gen_test
    async def test_MATERIALIZED_replace(self):
        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32,
            d String MATERIALIZED joinGet(other_database.other_table, 'val', toUInt32(b))
        """
        with self.assertRaisesRegex(tinybird.user.QueryNotAllowed, r"Resource 'other_database.other_table' not found"):
            await alter_table_operations(self.workspace, schema_a, schema_b)

        schema_c = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32,
            d String MATERIALIZED a IN (Select name from remote('127.0.0.1', 'system.tables'))
        """
        with self.assertRaisesRegex(tinybird.user.QueryNotAllowed, "DB::Exception: Usage of function remote is restricted"):
            await alter_table_operations(self.workspace, schema_a, schema_c)

    @tornado.testing.gen_test
    async def test_ALIAS_replace(self):
        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32,
            d String ALIAS joinGet(other_database.other_table, 'val', toUInt32(b))
        """
        with self.assertRaisesRegex(ValueError, r"ALIAS not supported"):
            await alter_table_operations(self.workspace, schema_a, schema_b)

        schema_c = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32,
            d String ALIAS a IN (Select name from remote('127.0.0.1', 'system.tables'))
        """
        with self.assertRaisesRegex(ValueError, r"ALIAS not supported"):
            await alter_table_operations(self.workspace, schema_a, schema_c)

    @tornado.testing.gen_test
    async def test_DEFAULT_replace(self):
        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32,
            d String DEFAULT joinGet(other_database.other_table, 'val', toUInt32(b))
        """
        with self.assertRaisesRegex(tinybird.user.QueryNotAllowed, r"Resource 'other_database.other_table' not found"):
            await alter_table_operations(self.workspace, schema_a, schema_b)

        schema_c = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32,
            d String DEFAULT a IN (Select name from remote('127.0.0.1', 'system.tables'))
        """
        with self.assertRaisesRegex(tinybird.user.QueryNotAllowed, "DB::Exception: Usage of function remote is restricted"):
            await alter_table_operations(self.workspace, schema_a, schema_c)

    @tornado.testing.gen_test
    async def test_add_several_columns(self):

        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32,
            d LowCardinality(String),
            e Int8 DEFAULT(0),
            f Nullable(Int32)
        """
        operations, operations_quarantine = await alter_table_operations(self.workspace, schema_a, schema_b)
        expected_ops = [
            'ADD COLUMN `d` LowCardinality(String)',
            'ADD COLUMN `e` Int8 DEFAULT 0',
            'ADD COLUMN `f` Nullable(Int32)',
        ]
        expected_quarantine_ops = [
            'ADD COLUMN `d` Nullable(String)',
            'ADD COLUMN `e` Nullable(String)',
            'ADD COLUMN `f` Nullable(String)',
        ]
        self.assertEqual(operations, expected_ops)
        self.assertEqual(operations_quarantine, expected_quarantine_ops)

    @tornado.testing.gen_test
    async def test_add_column_not_at_the_end(self):
        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = """
            aa Float32,
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        with self.assertRaisesRegex(ValueError, r"Could not add column 'aa' at position 1, new columns must be added at the end of the table"):
            await alter_table_operations(self.workspace, schema_a, schema_b)

    @tornado.testing.gen_test
    async def test_drop_column(self):
        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = """
            a String CODEC(LZ4HC(2)),
            b Float64
        """
        with self.assertRaisesRegex(ValueError, r"Dropping the 'c' column is not supported"):
            await alter_table_operations(self.workspace, schema_a, schema_b)

    @tornado.testing.gen_test
    async def test_add_and_drop_columns(self):
        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            d String
        """
        with self.assertRaisesRegex(ValueError, r"Dropping the 'c' column is not supported"):
            await alter_table_operations(self.workspace, schema_a, schema_b)

    @tornado.testing.gen_test
    async def test_modified_column(self):
        schema_a = """
            a String CODEC(LZ4HC(2)),
            b Float64,
            c Float32
        """
        schema_b = """
            a String CODEC(LZ4HC(2)),
            b String,
            c Float32
        """
        with self.assertRaisesRegex(ValueError, r"Modifying the 'b' column is not supported. Changing from '`b` Float64' to '`b` String'"):
            await alter_table_operations(self.workspace, schema_a, schema_b)




import unittest

from tinybird.csv_importer_find_endline_wrapper import find_new_line_index_to_split_the_buffer
from tinybird.csv_importer import BufferedCSVProcessor
from io import StringIO


class ChunkListCSVProcessor(BufferedCSVProcessor):
    def __init__(self, dialect):
        super().__init__(dialect)
        self.chunks = []

    def process_csv_chunk(self, csv_chunk):
        self.chunks.append(csv_chunk.decode())


class TestFindInCImplementation(unittest.TestCase):
    @staticmethod
    def _execute_find_new_line_index_to_split_the_buffer(buffer_text: str, quotechar='"', escapechar='\0'):

        buffer_text_encoded = buffer_text.encode("utf-8")
        idx = find_new_line_index_to_split_the_buffer(buffer_text_encoded, 'utf-8', quotechar, escapechar)
        print("IDX returned:", idx)
        process = buffer_text_encoded[:idx + 1]
        final_text = process.decode("utf-8")

        return final_text

    def test_buffering_logic_splits_the_csv_correctly_as_lines_come_in(self):

        csv = '''a,b,c
1,2,foo
4,5,bar
7,8,"foo

bar"
10,11,foo
13,14,bar
    '''
        csv_stream = StringIO(csv)

        p = ChunkListCSVProcessor(dialect={'new_line': "\n"})
        p.chunk_size = 16

        b = csv_stream.read(8)
        while b:
            p.write(b.encode())
            b = csv_stream.read(8)

        p.flush()

        for c in p.chunks:
            self.assertEqual(c.count('"') % 2, 0, f"{c} has invalid quoting")

    def test_buffering_logic_splits_the_csv_correctly_as_lines_come_in_with_fragment_bigger_than_chunk_size(self):

        csv = '''a,b,c
    1,2,foo
    4,5,bar
    7,8,"foo

    bar"
    10,11,foo
    13,14,bar
    '''
        csv_stream = StringIO(csv)

        p = ChunkListCSVProcessor(dialect={'escapechar': '"', 'new_line': "\n"})
        p.chunk_size = 16

        b = csv_stream.read(8)
        while b:
            p.write(b.encode())
            b = csv_stream.read(8)

        p.flush()

        for c in p.chunks:
            self.assertEqual(c.count('"') % 2, 0, f"{c} has invalid quoting")

    def buffering_fallbacks_to_old_new_line_finding_method_in_case_it_can_not_find_a_correct_new_line_char(self):

        csv = '''a,b,c
    1,2,"foo4,5,b
    ar7,8,foo10,11,foo11231231231233,14,bar
'''
        csv_stream = StringIO(csv)

        p = ChunkListCSVProcessor(dialect={'escapechar': '"', 'new_line': "\n"})
        p.chunk_size = 16

        b = csv_stream.read(8)

        while b:
            p.write(b.encode())
            b = csv_stream.read(8)
        p.flush()
        number_of_quotes_in_chunks = [c.count('"') for c in p.chunks]
        self.assertEqual([0, 1, 0], number_of_quotes_in_chunks)

    def test_split_on_complete_csv_returns_the_text(self):

        buffer_text = '''a,b,c
1,2,foo
4,5,bar
'''
        final_text = self._execute_find_new_line_index_to_split_the_buffer(buffer_text)

        self.assertEqual("""a,b,c
1,2,foo
4,5,bar
""", final_text)

    def test_split_csv_withlast_line_half_done_skips_that_line(self):

        buffer_text = '''a,b,c
1,2,foo
4,5,bar
7,8,"foo
'''
        final_text = self._execute_find_new_line_index_to_split_the_buffer(buffer_text)

        self.assertEqual("""a,b,c
1,2,foo
4,5,bar
""", final_text)

    def test_split_over_complete_csv_even_with_escaped_parts_returns_the_csv(self):

        buffer_text = '''a,b,c
1,2,foo
4,5,bar
7,8,"foo

bar"
10,11,foo
13,14,bar
'''
        final_text = self._execute_find_new_line_index_to_split_the_buffer(buffer_text)

        self.assertEqual('''a,b,c
1,2,foo
4,5,bar
7,8,"foo

bar"
10,11,foo
13,14,bar
''', final_text)

    def test_split_over_a_complete_csv_with_multiple_escaped_parts_returns_the_csv(self):

        buffer_text = '''a,b,c
1,2,"foo
"
4,"5",bar
7,"8","foo

bar"
10,11,foo
13,14,bar
'''
        final_text = self._execute_find_new_line_index_to_split_the_buffer(buffer_text)

        self.assertEqual('''a,b,c
1,2,"foo
"
4,"5",bar
7,"8","foo

bar"
10,11,foo
13,14,bar
''', final_text)

    def test_split_with_escaped_escapechar_and_and_incomplete_line_correctly_splits_the_text(self):

        buffer_text = '''a,b,c
1,2,""foo
4,""5"",bar
7,"""8","""foo


'''
        final_text = self._execute_find_new_line_index_to_split_the_buffer(buffer_text)

        self.assertEqual('''a,b,c
1,2,""foo
4,""5"",bar
''', final_text)

    def test_split_correctly_detects_and_endline_after_closing_a_quote(self):

        buffer_text = 'a,b,"c"\n4,"""5","bar"\n7,"8","foo\n'
        final_text = self._execute_find_new_line_index_to_split_the_buffer(buffer_text)
        self.assertEqual('a,b,"c"\n4,"""5","bar"\n', final_text)

    def test_buffer_using_multichar_new_line(self):

        buffer_text = 'a,b,c\r\n4,"5",bar\r\n7,"8","foo\r\n'
        final_text = self._execute_find_new_line_index_to_split_the_buffer(buffer_text)
        self.assertEqual('a,b,c\r\n4,"5",bar\r\n', final_text)

    def test_using_a_different_quotechar(self):
        buffer_text = "a,b,c\n4,'5',bar\n7,'8','foo\n"
        final_text = self._execute_find_new_line_index_to_split_the_buffer(buffer_text, quotechar="'")
        self.assertEqual("a,b,c\n4,'5',bar\n", final_text)

    def test_adding_a_different_kind_of_escapechar(self):
        buffer_text = 'a,b,c\n4,x"5x",bar\n7,"8x",x"foo\n'
        final_text = self._execute_find_new_line_index_to_split_the_buffer(buffer_text, escapechar='x')
        self.assertEqual('a,b,c\n4,x"5x",bar\n', final_text)

    def test_slow_iteration(self):
        csv = '''"Samsung U600""
    24""
    ","10000003409","1","10000003427"
"Samsung ""U600""
    24""","10000003409","",""
"Samsung U600 24""","10000003409","1","10000003427"
'''

        csv_stream = StringIO(csv)

        p = ChunkListCSVProcessor(dialect={'escapechar': '"', 'new_line': "\n"})
        p.chunk_size = 2

        b = csv_stream.read(1)
        while b:
            p.write(b.encode())
            b = csv_stream.read(1)

        p.flush()
        self.assertEqual(len(p.chunks), 3)



from tinybird.csv_importer_find_endline_wrapper import find_new_line_index_to_split_the_buffer
import unittest
import random


class TestCSVFindEndline(unittest.TestCase):
    def test_fuzzy(self):
        self.fuzzy(1)
        self.fuzzy(2)
        self.fuzzy(3)
        self.fuzzy(629574684)
        self.fuzzy(1054762918)
        self.fuzzy(629574684)
        i = 0
        while i < 2000:
            seed = random.randint(0, 2**31)
            self.fuzzy(seed)
            i += 1

    def fuzzy(self, seed):
        prng = random.Random(seed)
        text = ""

        FIELD, NEWLINE, QUOTED_FIELD, ESCAPED_NEWLINE, ESCAPED_QUOTE = 1, 2, 3, 4, 5
        iters = prng.randrange(0, 33)
        i = 0
        last_newline = -1
        last = None
        while i < iters:
            action = prng.sample([FIELD, NEWLINE, QUOTED_FIELD, ESCAPED_NEWLINE, ESCAPED_QUOTE], 1)[0]
            if last in {FIELD, ESCAPED_NEWLINE, ESCAPED_QUOTE, QUOTED_FIELD} and action != NEWLINE:
                text += ","
            if action == FIELD:
                size = prng.randrange(0, 33)
                text += "a" * size
            elif action == ESCAPED_NEWLINE:
                size = prng.randrange(0, 33)
                text += "\\\n" * size
            elif action == ESCAPED_QUOTE:
                size = prng.randrange(0, 33)
                text += '\\"' * size
            elif action == NEWLINE:
                last_newline = len(text)
                text += "\n"
            elif action == QUOTED_FIELD:
                size = prng.randrange(0, 33)
                t = "\n" * size
                text += f'"{t}"'
            last = action
            i += 1
        got = find_new_line_index_to_split_the_buffer(text.encode('utf-8'), 'utf-8', escapechar="\\")
        assert got == last_newline, f"Fuzzy test failed with seed {seed}: {repr(text)}"
